from __future__ import annotations

from dxtbx.format import Registry


def test_reading_refl_failure(dials_data):
    test_data = dials_data("centroid_test_data", pathlib=True)

    # Without dials_regression, none of the dials-data tests check for this "invalid binary data" case
    assert Registry.get_format_class_for_file(test_data / "indexed.refl") is None
    # Check .expt while here
    assert Registry.get_format_class_for_file(test_data / "indexed.expt") is None



"""Format classes to specifically recognise images from an electron detector
with a 2x2 array of Timepix modules, converted to SMV in various ways."""


from __future__ import annotations

import calendar
import sys
import time

from iotbx.detectors import SMVImage
from scitbx import matrix

from dxtbx.format.FormatSMV import FormatSMV
from dxtbx.model.beam import Probe
from dxtbx.model.detector import Detector


class FormatSMVTimePix_SU(FormatSMV):
    """Base format class to specifically recognise images from a Timepix-based
    detector, installed on a JEOL-2100 electron microscope at Stockholm
    University, which have been preprocessed by Wei Wan's RED software,
    to produce SMV files. See https://doi.org/10.1107/S0021889813027714 for RED.
    The detector itself has a 2x2 array of Timepix modules. Derived classes will
    either treat these separately as a multi-panel model, or as a single panel
    model."""

    @staticmethod
    def understand(image_file):

        size, header = FormatSMV.get_smv_header(image_file)

        # only recognise TimePix_SU
        if header.get("BEAMLINE", "").upper() != "TIMEPIX_SU":
            return False

        # check the header contains the things we're going to use
        wanted_header_items = [
            "BEAM_CENTER_X",
            "BEAM_CENTER_Y",
            "DISTANCE",
            "WAVELENGTH",
            "PIXEL_SIZE",
            "OSC_START",
            "OSC_RANGE",
            "PHI",
            "SIZE1",
            "SIZE2",
            "BYTE_ORDER",
            "DETECTOR_SN",
        ]
        if any(item not in header for item in wanted_header_items):
            return False

        return True

    def detectorbase_start(self):
        if not hasattr(self, "detectorbase") or self.detectorbase is None:
            self.detectorbase = SMVImage(self._image_file)
            self.detectorbase.open_file = self.open_file
            self.detectorbase.readHeader()

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. For this beamline
        this should be close to the provided values and neither aligned with the
        detector slow or fast axes."""

        return self._goniometer_factory.known_axis((-0.755, -0.656, 0.0))

    def _beam(self):
        """Return an unpolarized beam model."""

        wavelength = float(self._header_dictionary["WAVELENGTH"])

        return self._beam_factory.make_polarized_beam(
            sample_to_source=(0.0, 0.0, 1.0),
            wavelength=wavelength,
            polarization=(0, 1, 0),
            polarization_fraction=0.5,
            probe=Probe.electron,
        )

    def _scan(self):
        """Return the scan information for this image."""
        exposure_time = float(self._header_dictionary["TIME"])
        epoch = None

        # PST, PDT timezones not recognised by default...

        epoch = 0
        try:
            date_str = self._header_dictionary["DATE"]
            date_str = date_str.replace("PST", "").replace("PDT", "")
        except KeyError:
            date_str = ""
        for format_string in ["%a %b %d %H:%M:%S %Y", "%a %b %d %H:%M:%S %Z %Y"]:
            try:
                epoch = calendar.timegm(time.strptime(date_str, format_string))
                break
            except ValueError:
                pass

        # assert(epoch)
        osc_start = float(self._header_dictionary["OSC_START"])
        osc_range = float(self._header_dictionary["OSC_RANGE"])

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )


class FormatSMVTimePix_SU_512x512(FormatSMVTimePix_SU):
    """Implementation of a format class to specifically recognise images from a
    Timepix-based detector, installed on a JEOL-2100 electron microscope at
    Stockholm University, which have been preprocessed by Wei Wan's RED software,
    to produce SMV files. See https://doi.org/10.1107/S0021889813027714 for RED.
    The detector itself has a 2x2 array of Timepix modules."""

    @staticmethod
    def understand(image_file):

        size, header = FormatSMVTimePix_SU.get_smv_header(image_file)

        # check the pixel size is 55 microns
        if float(header["PIXEL_SIZE"]) != 0.055:
            return False

        # check there are 512*512 pixels
        if header["SIZE1"] != "512":
            return False
        if header["SIZE2"] != "512":
            return False

        return True

    def _detector(self):
        """4 panel detector, 55 micron pixels except for pixels at the outer
        edge of each chip, which are 165 microns wide."""
        # expect 55 mu pixels here, but make it general anyway
        pixel_size = tuple([float(self._header_dictionary["PIXEL_SIZE"])] * 2)
        image_size = (
            int(self._header_dictionary["SIZE1"]),
            int(self._header_dictionary["SIZE2"]),
        )
        panel_size = tuple([int(e / 2) for e in image_size])

        # outer pixels have three times the width
        # panel_size_mm = (
        #    pixel_size[0] * 3 + (panel_size[0] - 2) * pixel_size[0],
        #    pixel_size[1] * 3 + (panel_size[1] - 2) * pixel_size[1],
        # )
        trusted_range = (0, 65535)
        thickness = 0.3  # assume 300 mu thick

        # Initialise detector frame
        fast = matrix.col((1.0, 0.0, 0.0))
        slow = matrix.col((0.0, -1.0, 0.0))
        beam_centre = (
            float(self._header_dictionary["BEAM_CENTER_X"]),
            float(self._header_dictionary["BEAM_CENTER_Y"]),
        )

        bx_px, by_px = beam_centre

        def px_to_mm(px, px_size_1d, panel_size_1d):
            # the beam centre is in pixels. We want to convert to mm, taking the
            # different size of outer pixels into account. Use this local function
            # to do that
            mm = 0
            if px > 1:  # add first outer pixel
                mm += px_size_1d * 3
            else:  # or fraction of first outer pixel
                mm += px * px_size_1d * 3
                return mm

            if px > panel_size_1d - 1:  # add full panel of inner pixels
                mm += (panel_size_1d - 2) * px_size_1d
            else:  # or fraction of inner pixels
                mm += (px - 1) * px_size_1d
                return mm

            if px > panel_size_1d:  # add second outer pixel
                mm += px_size_1d * 3
            else:  # or fraction of second outer pixel
                mm += (px - (panel_size_1d - 1)) * px_size_1d * 3
                return mm

            if px > panel_size_1d + 1:  # add first outer pixel of second panel
                mm += px_size_1d * 3
            else:  # or fraction of first outer pixel of second panel
                mm += (px - panel_size_1d) * px_size_1d * 3
                return mm

            if px > (2 * panel_size_1d - 1):  # add second full panel of inner pixels
                mm += (panel_size_1d - 2) * px_size_1d
                # plus remaining fraction of the second outer pixel
                mm += (px - (2 * panel_size_1d - 1)) * px_size_1d * 3
            else:  # or fraction of inner pixels of the second panel
                mm += (px - panel_size_1d - 1) * px_size_1d
            return mm

        bx_mm = px_to_mm(bx_px, pixel_size[0], panel_size[0])
        by_mm = px_to_mm(by_px, pixel_size[1], panel_size[1])

        # the beam centre is defined from the origin along fast, slow. To determine
        # the lab frame origin we place the beam centre down the -z axis
        dist = float(self._header_dictionary["DISTANCE"])
        cntr = matrix.col((0.0, 0.0, -1 * dist))
        orig = cntr - bx_mm * fast - by_mm * slow

        d = Detector()

        root = d.hierarchy()
        root.set_local_frame(fast.elems, slow.elems, orig.elems)

        self.coords = {}
        panel_idx = 0

        # set panel extent in pixel numbers and x, y mm shifts. Note that the
        # outer pixels are 0.165 mm in size. These are excluded from the panel
        # extents.
        pnl_data = []
        pnl_data.append(
            {
                "xmin": 1,
                "ymin": 1,
                "xmax": 255,
                "ymax": 255,
                "xmin_mm": 1 * 0.165,
                "ymin_mm": 1 * 0.165,
            }
        )
        pnl_data.append(
            {
                "xmin": 257,
                "ymin": 1,
                "xmax": 511,
                "ymax": 255,
                "xmin_mm": 3 * 0.165 + (511 - 257) * pixel_size[0],
                "ymin_mm": 1 * 0.165,
            }
        )
        pnl_data.append(
            {
                "xmin": 1,
                "ymin": 257,
                "xmax": 255,
                "ymax": 511,
                "xmin_mm": 1 * 0.165,
                "ymin_mm": 3 * 0.165 + (511 - 257) * pixel_size[1],
            }
        )
        pnl_data.append(
            {
                "xmin": 257,
                "ymin": 257,
                "xmax": 511,
                "ymax": 511,
                "xmin_mm": 3 * 0.165 + (511 - 257) * pixel_size[0],
                "ymin_mm": 3 * 0.165 + (511 - 257) * pixel_size[1],
            }
        )

        # redefine fast, slow for the local frame
        fast = matrix.col((1.0, 0.0, 0.0))
        slow = matrix.col((0.0, 1.0, 0.0))

        for ipanel, pd in enumerate(pnl_data):
            xmin = pd["xmin"]
            xmax = pd["xmax"]
            ymin = pd["ymin"]
            ymax = pd["ymax"]
            xmin_mm = pd["xmin_mm"]
            ymin_mm = pd["ymin_mm"]

            origin_panel = fast * xmin_mm + slow * ymin_mm

            panel_name = "Panel%d" % panel_idx
            panel_idx += 1

            p = d.add_panel()
            p.set_type("SENSOR_PAD")
            p.set_name(panel_name)
            p.set_raw_image_offset((xmin, ymin))
            p.set_image_size((xmax - xmin, ymax - ymin))
            p.set_trusted_range(trusted_range)
            p.set_pixel_size((pixel_size[0], pixel_size[1]))
            p.set_thickness(thickness)
            p.set_material("Si")
            # p.set_mu(mu)
            # p.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, t0))
            p.set_local_frame(fast.elems, slow.elems, origin_panel.elems)
            p.set_raw_image_offset((xmin, ymin))
            self.coords[panel_name] = (xmin, ymin, xmax, ymax)

        return d

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        raw_data = self._get_endianic_raw_data(size=(512, 512))

        # split into separate panels
        self._raw_data = []
        d = self.get_detector()
        for panel in d:
            xmin, ymin, xmax, ymax = self.coords[panel.get_name()]
            self._raw_data.append(raw_data[ymin:ymax, xmin:xmax])

        return tuple(self._raw_data)


class FormatSMVTimePix_SU_516x516(FormatSMVTimePix_SU):
    """A class for reading SMV format images for the Timepix-based electron
    detector where the wider edge pixels have been split into three normal-sized
    pixels. The whole 516*516 detector can then be described as a single panel"""

    @staticmethod
    def understand(image_file):

        size, header = FormatSMVTimePix_SU.get_smv_header(image_file)

        # check there are 516*516 pixels
        if header["SIZE1"] != "516":
            return False
        if header["SIZE2"] != "516":
            return False

        return True

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. For this beamline
        this should be close to the provided values and neither aligned with the
        detector slow or fast axes."""

        # import numpy as np
        # angle = np.radians(-40) #-0.672  # radians
        # axis = np.cos(angle), -np.cos(angle+np.pi/2), 0

        # axis = (0.755, -0.656, 0.0)
        axis = (0.7826, -0.6226, 0.0)
        return self._goniometer_factory.known_axis(axis)

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        distance = float(self._header_dictionary["DISTANCE"])  # mm

        beam_x = float(self._header_dictionary["BEAM_CENTER_X"])  # px
        beam_y = float(self._header_dictionary["BEAM_CENTER_Y"])  # px

        # wavelength = float(self._header_dictionary['WAVELENGTH'])  # Angstrom

        pixelsize = float(self._header_dictionary["PIXEL_SIZE"])  # mm

        thickness = 0.3  # mm
        material = "Si"

        nx = int(self._header_dictionary["SIZE1"])  # number of pixels
        ny = int(self._header_dictionary["SIZE2"])  # number of pixels

        underload, overload = (-1, 65535)

        detector = self._detector_factory.simple(
            sensor="PAD",
            distance=distance,
            beam_centre=(beam_x * pixelsize, beam_y * pixelsize),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(pixelsize, pixelsize),
            image_size=(nx, ny),
            trusted_range=(underload, overload),
        )

        detector[0].set_thickness(thickness)
        detector[0].set_material(material)
        detector[0].mask = []  # can we mask the cross here?

        return detector

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        nx = int(self._header_dictionary["SIZE1"])  # number of pixels
        ny = int(self._header_dictionary["SIZE2"])  # number of pixels
        raw_data = self._get_endianic_raw_data(size=(nx, ny))

        self._raw_data = raw_data

        return self._raw_data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print("FormatSMVTimePix_SU:", FormatSMVTimePix_SU.understand(arg))
        print(
            "FormatSMVTimePix_SU_512x512:", FormatSMVTimePix_SU_512x512.understand(arg)
        )
        print(
            "FormatSMVTimePix_SU_516x516:", FormatSMVTimePix_SU_516x516.understand(arg)
        )



from __future__ import annotations

import sys

from libtbx.phil import parse

from dxtbx.format.FormatXTC import FormatXTC, locator_str
from dxtbx.format.FormatXTCCspad import FormatXTCCspad, cspad_locator_str
from dxtbx.format.FormatXTCJungfrau import FormatXTCJungfrau, jungfrau_locator_str
from dxtbx.format.FormatXTCRayonix import FormatXTCRayonix, rayonix_locator_str
from dxtbx.model import Detector

multiple_locator_scope = parse(
    rayonix_locator_str + cspad_locator_str + jungfrau_locator_str + locator_str,
    process_includes=True,
)


class FormatXTCMultipleDetectors(FormatXTCRayonix, FormatXTCCspad, FormatXTCJungfrau):
    def __init__(self, image_file, **kwargs):
        if any(["rayonix" in src.lower() for src in self.params.detector_address]):
            FormatXTCRayonix.__init__(self, image_file, **kwargs)
        if any(["cspad" in src.lower() for src in self.params.detector_address]):
            FormatXTCCspad.__init__(self, image_file, **kwargs)
        if any(["jungfrau" in src.lower() for src in self.params.detector_address]):
            FormatXTCJungfrau.__init__(self, image_file, **kwargs)
        FormatXTC.__init__(
            self, image_file, locator_scope=multiple_locator_scope, **kwargs
        )

    @staticmethod
    def understand(image_file):
        try:
            params = FormatXTC.params_from_phil(multiple_locator_scope, image_file)
        except Exception:
            return False

        if params.detector_address is None or len(params.detector_address) <= 1:
            return False
        return [
            "rayonix" in src.lower()
            or "cspad" in src.lower()
            or "jungfrau" in src.lower()
            for src in params.detector_address
        ].count(True) >= 2

    def get_raw_data(self, index=None):
        if index is None:
            index = 0

        all_addresses = self.params.detector_address

        raw_data = []

        for address in all_addresses:
            self.params.detector_address = [address]
            if "rayonix" in address.lower():
                data = FormatXTCRayonix.get_raw_data(self, index)
            elif "cspad" in address.lower():
                data = FormatXTCCspad.get_raw_data(self, index)
            elif "jungfrau" in address.lower():
                data = FormatXTCJungfrau.get_raw_data(self, index)
            assert data is not None, address
            if not isinstance(data, tuple):
                data = (data,)
            raw_data.extend(data)
        self.params.detector_address = all_addresses

        self._raw_data = raw_data
        return tuple(raw_data)

    def get_detector(self, index=None):
        return self._detector(index)

    def _detector(self, index=None):
        if index is None:
            index = 0
        d = Detector()
        root = d.hierarchy()

        all_addresses = self.params.detector_address

        def recursive_add_node(a, b):
            # add a to b
            if a.is_panel():
                b.add_panel(a)
            else:
                g = b.add_group(a)
                for child in a:
                    recursive_add_node(child, g)

        for address in all_addresses:
            self.params.detector_address = [address]
            sub_d = None
            try:
                if "rayonix" in address.lower():
                    sub_d = FormatXTCRayonix._detector(self)
                elif "cspad" in address.lower():
                    sub_d = FormatXTCCspad._detector(self, index)
                elif "jungfrau" in address.lower():
                    sub_d = FormatXTCJungfrau._detector(self, index)
            except Exception:
                continue
            assert sub_d is not None, address
            recursive_add_node(sub_d.hierarchy(), root)
        self.params.detector_address = all_addresses

        return d


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatXTCMultipleDetectors.understand(arg))



from __future__ import annotations

import os
import sys

import numpy as np

try:
    import psana
except ImportError:
    psana = None
except TypeError:
    # Check if SIT_* environment variables are set
    if os.environ.get("SIT_ROOT"):
        # Variables are present, so must have been another error
        raise
    psana = None

from cctbx import factor_kev_angstrom
from cctbx.eltbx import attenuation_coefficient
from libtbx.phil import parse
from scitbx.array_family import flex
from scitbx.matrix import col

from dxtbx.format.FormatXTC import FormatXTC, locator_str
from dxtbx.model import Detector, ParallaxCorrectedPxMmStrategy

jungfrau_locator_str = """
  jungfrau {
    dark = True
      .type = bool
      .help = Dictates if dark subtraction is done from raw data
    monolithic = False
      .type = bool
      .help = switch to FormatXTCJungfrauMonolithic if True. Used for LS49 image averaging
    use_big_pixels = True
      .type = bool
      .help = account for multi-sized pixels in the 512x1024 Jungfrau panels, forming a 514x1030 pixel panel
    detz_offset = None
      .type = float
      .help = Distance from back of detector rail to sample interaction region (CXI) \
              or actual detector distance (XPP/MFX)
  }
"""

jungfrau_locator_scope = parse(
    jungfrau_locator_str + locator_str, process_includes=True
)


class FormatXTCJungfrau(FormatXTC):
    def __init__(self, image_file, **kwargs):
        super().__init__(image_file, locator_scope=jungfrau_locator_scope, **kwargs)
        self._cached_detector = {}

        self._dist_det = None

    @staticmethod
    def understand(image_file):
        try:
            params = FormatXTC.params_from_phil(jungfrau_locator_scope, image_file)
        except Exception:
            return False
        return any(["jungfrau" in src.lower() for src in params.detector_address])

    def get_raw_data(self, index=None):
        from serialtbx.detector.util import jungfrau

        if index is None:
            index = 0

        d = FormatXTCJungfrau.get_detector(self, index)
        evt = self._get_event(index)
        run = self.get_run_from_index(index)
        det = self._get_psana_detector(run)
        data = det.calib(evt)
        data = data.astype(np.float64)
        self._raw_data = []
        for module_count, module in enumerate(d.hierarchy()):

            if (
                self.params.jungfrau.use_big_pixels
                and os.environ.get("DONT_USE_BIG_PIXELS_JUNGFRAU") is None
            ):
                panel_data = jungfrau.correct_panel(data[module_count])
                self._raw_data.append(flex.double(panel_data))
                continue

            for asic_count, asic in enumerate(module):
                fdim, sdim = asic.get_image_size()
                sensor_id = asic_count // 4  # There are 2X4 asics per module
                asic_in_sensor_id = asic_count % 4  # this number will be 0,1,2 or 3
                asic_data = data[module_count][
                    sensor_id * sdim : (sensor_id + 1) * sdim,
                    asic_in_sensor_id * fdim : (asic_in_sensor_id + 1) * fdim,
                ]  # 8 sensors per module
                self._raw_data.append(flex.double(np.array(asic_data)))
        assert len(d) == len(self._raw_data), (len(d), len(self._raw_data))
        return tuple(self._raw_data)

    def get_detector(self, index=None):
        return FormatXTCJungfrau._detector(self, index)

    def _detector(self, index=None):
        from PSCalib.SegGeometryStore import sgs
        from serialtbx.detector.xtc import basis_from_geo

        run = self.get_run_from_index(index)
        if run.run() in self._cached_detector:
            return self._cached_detector[run.run()]

        if index is None:
            index = 0
        assert len(self.params.detector_address) == 1
        self._det = psana.Detector(self.params.detector_address[0], run.env())
        evt = self._get_event(index)
        wavelength = self.get_beam(index).get_wavelength()

        if self._dist_det is None:
            self._dist_det = psana.Detector("CXI:DS1:MMS:06.RBV")

        geom = self._det.pyda.geoaccess(evt.run())
        pixel_size = (
            self._det.pixel_size(self._get_event(index)) / 1000.0
        )  # convert to mm
        d = Detector()
        pg0 = d.hierarchy()
        # first deal with D0
        det_num = 0
        root = geom.get_top_geo()
        root_basis = basis_from_geo(root)
        while len(root.get_list_of_children()) == 1:
            sub = root.get_list_of_children()[0]
            sub_basis = basis_from_geo(sub)
            root = sub
            root_basis = root_basis * sub_basis
        t = root_basis.translation
        if self.params.jungfrau.detz_offset:
            distance = self._dist_det(evt) + self.params.jungfrau.detz_offset
        else:
            distance = t[2]
        root_basis.translation = col((t[0], t[1], -distance))

        origin = col((root_basis * col((0, 0, 0, 1)))[0:3])
        fast = col((root_basis * col((1, 0, 0, 1)))[0:3]) - origin
        slow = col((root_basis * col((0, 1, 0, 1)))[0:3]) - origin
        pg0.set_local_frame(fast.elems, slow.elems, origin.elems)
        pg0.set_name("D%d" % (det_num))

        # Now deal with modules
        for module_num in range(len(root.get_list_of_children())):
            module = root.get_list_of_children()[module_num]
            module_basis = basis_from_geo(module)
            origin = col((module_basis * col((0, 0, 0, 1)))[0:3])
            fast = col((module_basis * col((1, 0, 0, 1)))[0:3]) - origin
            slow = col((module_basis * col((0, 1, 0, 1)))[0:3]) - origin
            pg1 = pg0.add_group()
            pg1.set_local_frame(fast.elems, slow.elems, origin.elems)
            pg1.set_name("D%dM%d" % (det_num, module_num))

            # Read the known layout of the Jungfrau 2x4 module
            sg = sgs.Create(segname=module.oname)
            xx, yy = sg.get_seg_xy_maps_um()
            xx = xx / 1000
            yy = yy / 1000

            # Now deal with ASICs
            for asic_num in range(8):
                val = "ARRAY_D0M%dA%d" % (module_num, asic_num)
                dim_slow = xx.shape[0]
                dim_fast = xx.shape[1]
                sensor_id = asic_num // 4  # There are 2X4 asics per module
                asic_in_sensor_id = asic_num % 4  # this number will be 0,1,2 or 3
                id_slow = sensor_id * (dim_slow // 2)
                id_fast = asic_in_sensor_id * (dim_fast // 4)
                origin = col((xx[id_slow][id_fast], yy[id_slow][id_fast], 0))
                fp = col((xx[id_slow][id_fast + 1], yy[id_slow][id_fast + 1], 0))
                sp = col((xx[id_slow + 1][id_fast], yy[id_slow + 1][id_fast], 0))
                fast = (fp - origin).normalize()
                slow = (sp - origin).normalize()
                p = pg1.add_panel()
                p.set_local_frame(fast.elems, slow.elems, origin.elems)
                p.set_pixel_size((pixel_size, pixel_size))
                p.set_trusted_range((-10, 2e6))
                p.set_name(val)

                thickness, material = 0.32, "Si"
                p.set_thickness(thickness)  # mm
                p.set_material(material)
                # Compute the attenuation coefficient.
                # This will fail for undefined composite materials
                # mu_at_angstrom returns cm^-1, but need mu in mm^-1
                table = attenuation_coefficient.get_table(material)
                mu = table.mu_at_angstrom(wavelength) / 10.0
                p.set_mu(mu)
                p.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, thickness))
                p.set_gain(factor_kev_angstrom / wavelength)

                if (
                    self.params.jungfrau.use_big_pixels
                    and os.environ.get("DONT_USE_BIG_PIXELS_JUNGFRAU") is None
                ):
                    p.set_image_size((1030, 514))
                    break
                else:
                    p.set_image_size((dim_fast // 4, dim_slow // 2))
        if (
            self.params.jungfrau.use_big_pixels
            and os.environ.get("DONT_USE_BIG_PIXELS_JUNGFRAU") is None
        ):
            assert len(d) == 8
        self._cached_detector[run.run()] = d
        return d


class FormatXTCJungfrauMonolithic(FormatXTCJungfrau):
    """Monolithic version of the Jungfrau, I.E. use the psana detector image function to assemble a monolithic image"""

    @staticmethod
    def understand(image_file):
        try:
            params = FormatXTC.params_from_phil(jungfrau_locator_scope, image_file)
            if params.jungfrau.monolithic:
                return True
            return False
        except Exception:
            return False

    def get_raw_data(self, index=None):
        if index is None:
            index = 0
        self.get_detector(index)  # is this line required?
        evt = self._get_event(index)
        run = self.get_run_from_index(index)
        det = self._get_psana_detector(run)
        data = det.image(evt)
        data = data.astype(np.float64)
        self._raw_data = flex.double(data)
        return self._raw_data

    def get_detector(self, index=None):
        return self._detector(index)

    def _detector(self, index=None):
        return self._detector_factory.simple(
            sensor="UNKNOWN",
            distance=100.0,
            beam_centre=(50.0, 50.0),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(0.075, 0.075),
            image_size=(1030, 1064),
            trusted_range=(-9, 2e6),
            mask=[],
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        # Bug, should call this part differently for understand method to work
        print(FormatXTCJungfrau.understand(arg))



from __future__ import annotations

import string

import h5py
import numpy as np

from cctbx.eltbx import attenuation_coefficient
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.format.FormatNexus import FormatNexus
from dxtbx.format.nexus import (
    BeamFactory,
    DataFactory,
    MaskFactory,
    NXmxReader,
    convert_units,
)
from dxtbx.model import Detector, Panel, ParallaxCorrectedPxMmStrategy, Scan


def clean_string(input):
    if hasattr(string, "letters"):
        letters = string.letters
    else:
        letters = string.ascii_letters
    return "".join(i for i in input if i in letters)


class FormatNexusJungfrauHack(FormatNexus):
    @staticmethod
    def understand(image_file):
        try:
            with h5py.File(image_file, "r") as handle:
                return "/entry/instrument/JF1M" in handle
        except OSError:
            return False

    def _start(self):

        # Read the file structure
        self._reader = reader = NXmxReader(self._image_file)

        # Only support 1 set of models at the moment
        assert len(reader.entries) == 1, "Currently only supports 1 NXmx entry"
        assert len(reader.entries[0].data) == 1, "Currently only supports 1 NXdata"
        assert (
            len(reader.entries[0].instruments) == 1
        ), "Currently only supports 1 NXinstrument"
        assert len(reader.entries[0].samples) == 1, "Currently only supports 1 NXsample"
        assert (
            len(reader.entries[0].samples[0].beams) == 1
            or len(reader.entries[0].instruments[0].beams) == 1
        ), "Currently only supports 1 NXbeam"

        # Get the NXmx model objects
        entry = reader.entries[0]
        self.instrument = instrument = entry.instruments[0]
        detector = instrument.detectors[0]
        sample = entry.samples[0]
        beam = sample.beams[0] if sample.beams else instrument.beams[0]
        data = entry.data[0]

        # Construct the models
        self._beam_factory = BeamFactory(beam)
        self._beam_factory.load_model(0)

        self._setup_detector(detector, self._beam_factory.model)
        self._setup_gonio_and_scan(sample, detector)

        if self._scan_model:
            array_range = self._scan_model.get_array_range()
            num_images = array_range[1] - array_range[0]
        else:
            num_images = 0

        self._raw_data = DataFactory(data, max_size=num_images)

    def _setup_detector(self, detector, beam):
        nx_detector = detector.handle
        nx_module = detector.modules[0].handle

        # Get the detector name and type
        if "type" in nx_detector:
            detector_type = str(nx_detector["type"][()])
        else:
            detector_type = "unknown"
        detector_name = str(nx_detector.name)

        # Get the trusted range of pixel values
        if "saturation_value" in nx_detector:
            trusted_range = (0, float(nx_detector["saturation_value"][()]))
        else:
            trusted_range = (0, 99999999)

        # Get the detector thickness
        thickness = nx_detector["sensor_thickness"]
        thickness_value = float(thickness[()])
        thickness_units = thickness.attrs["units"]
        thickness_value = float(convert_units(thickness_value, thickness_units, "mm"))

        # Get the detector material
        material = nx_detector["sensor_material"][()]
        if hasattr(material, "shape"):
            material = "".join(m.decode() for m in material)
        detector_material = clean_string(str(material))
        material = {
            "Si": "Si",
            np.string_("Si"): "Si",
            np.string_("Silicon"): "Si",
            np.string_("Sillicon"): "Si",
            np.string_("CdTe"): "CdTe",
            np.string_("GaAs"): "GaAs",
        }.get(detector_material)
        if not material:
            raise RuntimeError("Unknown material: %s" % detector_material)

        try:
            x_pixel = nx_detector["x_pixel_size"][()] * 1000.0
            y_pixel = nx_detector["y_pixel_size"][()] * 1000.0

            legacy_beam_x = float(x_pixel * nx_detector["beam_center_x"][()])
            legacy_beam_y = float(y_pixel * nx_detector["beam_center_y"][()])
            legacy_distance = float(1000 * nx_detector["detector_distance"][()])
        except KeyError:
            legacy_beam_x = 0
            legacy_beam_y = 0

        # Get the fast pixel size and vector
        fast_pixel_direction = nx_module["fast_pixel_direction"]
        fast_pixel_direction_value = float(fast_pixel_direction[()])
        fast_pixel_direction_units = fast_pixel_direction.attrs["units"]
        fast_pixel_direction_vector = fast_pixel_direction.attrs["vector"]
        fast_pixel_direction_value = convert_units(
            fast_pixel_direction_value, fast_pixel_direction_units, "mm"
        )
        fast_axis = matrix.col(fast_pixel_direction_vector).normalize()

        # Get the slow pixel size and vector
        slow_pixel_direction = nx_module["slow_pixel_direction"]
        slow_pixel_direction_value = float(slow_pixel_direction[()])
        slow_pixel_direction_units = slow_pixel_direction.attrs["units"]
        slow_pixel_direction_vector = slow_pixel_direction.attrs["vector"]
        slow_pixel_direction_value = convert_units(
            slow_pixel_direction_value, slow_pixel_direction_units, "mm"
        )
        slow_axis = matrix.col(slow_pixel_direction_vector).normalize()

        origin = matrix.col((0.0, 0.0, -legacy_distance))
        if origin.elems[0] == 0.0 and origin.elems[1] == 0:
            origin = -(origin + legacy_beam_x * fast_axis + legacy_beam_y * slow_axis)

        # Change of basis to convert from NeXus to IUCr/ImageCIF convention
        cob = matrix.sqr((-1, 0, 0, 0, 1, 0, 0, 0, -1))
        origin = cob * matrix.col(origin)
        fast_axis = cob * fast_axis
        slow_axis = cob * slow_axis

        # Ensure that fast and slow axis are orthogonal
        normal = fast_axis.cross(slow_axis)
        slow_axis = -fast_axis.cross(normal)

        # Compute the attenuation coefficient.
        # This will fail for undefined composite materials
        # mu_at_angstrom returns cm^-1, but need mu in mm^-1
        table = attenuation_coefficient.get_table(material)
        wavelength = beam.get_wavelength()
        mu = table.mu_at_angstrom(wavelength) / 10.0

        # Construct the detector model
        pixel_size = (fast_pixel_direction_value, slow_pixel_direction_value)
        # image size stored slow to fast but dxtbx needs fast to slow
        image_size = tuple(int(x) for x in reversed(nx_module["data_size"][-2:]))
        image_size = (1030, 1064)

        self.model = Detector()
        self.model.add_panel(
            Panel(
                detector_type,
                detector_name,
                tuple(fast_axis),
                tuple(slow_axis),
                tuple(origin),
                pixel_size,
                image_size,
                trusted_range,
                thickness_value,
                material,
                mu,
            )
        )

        # Set the parallax correction
        for panel in self.model:
            panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, thickness_value))
            panel.set_type("SENSOR_PAD")

        self._detector_model = self.model

    def _setup_gonio_and_scan(self, sample, detector):
        with h5py.File(self._image_file, "r") as handle:
            phi = handle["/entry/sample/goniometer/omega"][()]
        image_range = (1, len(phi))
        oscillation = (float(phi[0]), float(phi[1] - phi[0]))

        # Get the exposure time
        num_images = len(phi)
        exposure_time = flex.double(num_images, 0)
        epochs = flex.double(num_images, 0.0)
        for i in range(1, len(epochs)):
            epochs[i] = epochs[i - 1] + exposure_time[i - 1]

        self._scan_model = Scan(image_range, oscillation, exposure_time, epochs)
        self._goniometer_model = self._goniometer_factory.known_axis((-1, 0, 0))

    def _end(self):
        return

    def _goniometer(self):
        return self._goniometer_model

    def _detector(self):
        return self._detector_model

    def _scan(self):
        return self._scan_model

    def get_goniometer(self, index=None):
        return self._goniometer()

    def get_detector(self, index=None):
        return self._detector()

    def get_beam(self, index=None):
        return self._beam(index)

    def get_scan(self, index=None):
        if index is None:
            return self._scan()
        scan = self._scan()
        if scan is not None:
            return scan[index]
        return scan

    def get_raw_data(self, index):
        return self._raw_data[index]

    def get_static_mask(self, index=None, goniometer=None):
        return MaskFactory(self.instrument.detectors, index).mask

    def get_num_images(self):
        if self._scan() is not None:
            return self._scan().get_num_images()
        return len(self._raw_data)

    def get_image_file(self, index=None):
        return self._image_file

    def get_detectorbase(self, index=None):
        raise NotImplementedError

    @staticmethod
    def get_instrument_name(handle):
        if "short_name" in handle["/entry/instrument"].attrs:
            name = handle["/entry/instrument"].attrs["short_name"]
        elif "/entry/instrument/name" in handle:
            if "short_name" in handle["/entry/instrument/name"].attrs:
                name = handle["/entry/instrument/name"].attrs["short_name"]
            else:
                name = handle["/entry/instrument/name"][()]
        else:
            name = None
        return name



from __future__ import annotations

import sys
from io import StringIO
from math import pi, sqrt

from cctbx.crystal_orientation import crystal_orientation
from cctbx.sgtbx import space_group as SG
from cctbx.sgtbx import space_group_symbols
from cctbx.uctbx import unit_cell
from scitbx import matrix
from scitbx.math import angle_derivative_wrt_vectors
from scitbx.math.lefebvre import matrix_inverse_error_propagation


class crystal_model_old:
    """Simple model for the crystal lattice geometry and symmetry

    A crystal is initialised from the elements of its real space axes
    a, b, and c. Space group information must also be provided, either
    in the form of a symbol, or an existing
    cctbx.sgtbx.space_group object. If space_group_symbol is provided,
    it is passed to the cctbx.sgtbx.space_group_symbols constructor.
    This accepts either extended Hermann Mauguin format, or Hall format
    with the prefix 'Hall:'. E.g.

    space_group_symbol = "P b a n:1"
        or
    space_group_symbol = "Hall:P 2 2 -1ab"

    Optionally the crystal mosaicity value may be set, with the deg
    parameter controlling whether this value is treated as being an
    angle in degrees or radians."""

    def __init__(
        self,
        real_space_a,
        real_space_b,
        real_space_c,
        space_group_symbol=None,
        space_group=None,
        mosaicity=None,
        deg=True,
    ):

        # Set the space group
        assert [space_group_symbol, space_group].count(None) == 1
        if space_group_symbol:
            self._sg = SG(space_group_symbols(space_group_symbol))
        else:
            self._sg = space_group

        # Set the mosaicity
        if mosaicity is not None:
            self.set_mosaicity(mosaicity, deg=deg)
        else:
            self._mosaicity = 0.0

        # setting matrix at initialisation
        real_space_a = matrix.col(real_space_a)
        real_space_b = matrix.col(real_space_b)
        real_space_c = matrix.col(real_space_c)
        A = matrix.sqr(
            real_space_a.elems + real_space_b.elems + real_space_c.elems
        ).inverse()

        # unit cell
        self.set_unit_cell(real_space_a, real_space_b, real_space_c)

        # reciprocal space orthogonalisation matrix, i.e. the transpose of the
        # real space fractionalisation matrix.
        # see lecture notes from Kevin Cowtan, 'Coordinate systems, operators,
        # and transformations'
        #  https://www.iucr.org/resources/commissions/computing/schools/siena-2005-crystallographic-computing-school/speakers-notes
        self._update_B()

        # initial orientation matrix
        self._U = A * self._B.inverse()

        # set up attributes for scan-varying model
        self.reset_scan_points()

        # set up attributes for unit cell errors
        self.reset_unit_cell_errors()

        return

    @property
    def num_scan_points(self):
        return self._num_scan_points

    def show(self, show_scan_varying=False, out=None):
        if out is None:
            out = sys.stdout
        uc = self.get_unit_cell().parameters()
        sg = str(self.get_space_group().info())
        umat = (
            self.get_U()
            .mathematica_form(format="% 5.4f", one_row_per_line=True)
            .splitlines()
        )
        bmat = (
            self.get_B()
            .mathematica_form(format="% 5.4f", one_row_per_line=True)
            .splitlines()
        )
        amat = (
            (self.get_U() * self.get_B())
            .mathematica_form(format="% 5.4f", one_row_per_line=True)
            .splitlines()
        )

        msg = ["Crystal:"]
        msg.append(
            "    Unit cell: " + "(%5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f)" % uc
        )
        msg.append("    Space group: " + sg)
        msg.append("    U matrix:  " + umat[0])
        msg.append("               " + umat[1])
        msg.append("               " + umat[2])
        msg.append("    B matrix:  " + bmat[0])
        msg.append("               " + bmat[1])
        msg.append("               " + bmat[2])
        msg.append("    A = UB:    " + amat[0])
        msg.append("               " + amat[1])
        msg.append("               " + amat[2])
        if self.num_scan_points > 0:
            msg.append("    A sampled at " + str(self.num_scan_points) + " scan points")
            if show_scan_varying:
                for i in range(self.num_scan_points):
                    A = self.get_A_at_scan_point(i)
                    B = self.get_B_at_scan_point(i)
                    U = self.get_U_at_scan_point(i)
                    uc = self.get_unit_cell_at_scan_point(i).parameters()
                    umat = U.mathematica_form(
                        format="% 5.4f", one_row_per_line=True
                    ).splitlines()
                    bmat = B.mathematica_form(
                        format="% 5.4f", one_row_per_line=True
                    ).splitlines()
                    amat = A.mathematica_form(
                        format="% 5.4f", one_row_per_line=True
                    ).splitlines()
                    msg.append("  Scan point #%i:" % (i + 1))
                    msg.append(
                        "    Unit cell: "
                        + "(%5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f)" % uc
                    )
                    msg.append("    U matrix:  " + umat[0])
                    msg.append("               " + umat[1])
                    msg.append("               " + umat[2])
                    msg.append("    B matrix:  " + bmat[0])
                    msg.append("               " + bmat[1])
                    msg.append("               " + bmat[2])
                    msg.append("    A = UB:    " + amat[0])
                    msg.append("               " + amat[1])
                    msg.append("               " + amat[2])
        print("\n".join(msg), file=out)

    def __str__(self):
        s = StringIO()
        self.show(out=s)
        return s.getvalue()

    def set_unit_cell(self, real_space_a, real_space_b, real_space_c):
        cell = (
            real_space_a.length(),
            real_space_b.length(),
            real_space_c.length(),
            real_space_b.angle(real_space_c, deg=True),
            real_space_c.angle(real_space_a, deg=True),
            real_space_a.angle(real_space_b, deg=True),
        )
        self._uc = unit_cell(cell)
        self._update_B()

    def get_unit_cell_at_scan_point(self, t):
        B = self.get_B_at_scan_point(t)
        co = crystal_orientation(B)
        return co.unit_cell()

    def _update_B(self):
        self._B = matrix.sqr(self._uc.fractionalization_matrix()).transpose()

    def set_U(self, U):

        # check U is a rotation matrix.
        assert U.is_r3_rotation_matrix()
        self._U = U

        # reset scan-varying data, if the static U has changed
        self.reset_scan_points()

    def get_U(self):
        return self._U

    def get_B(self):
        return self._B

    def set_B(self, B):

        # also set the unit cell
        co = crystal_orientation(B, True)
        self._uc = co.unit_cell()
        self._B = matrix.sqr(self._uc.fractionalization_matrix()).transpose()

        # reset scan-varying data, if the static B has changed
        self.reset_scan_points()

        # reset unit cell errors
        self.reset_unit_cell_errors()

    def get_B_covariance(self):
        """Return the 9*9 covariance matrix of elements of B, if available,
        otherwise None. The order of elements of this matrix are determined by
        'flattening' B to form a 1D vector using row major ordering.

        :rtype: :py:class:`scitbx.matrix.sqr`
        """
        return self._cov_B

    def set_B_covariance(self, cov):

        cov = matrix.sqr(cov)

        # check cov is of the right size. No other checks made
        assert cov.n == (9, 9)

        self._cov_B = cov

        return

    def _calc_cell_parameter_sd(self):

        # self._cov_B is the covariance matrix of elements of the B matrix. We
        # need to construct the covariance matrix of elements of the
        # transpose of B. The vector of elements of B is related to the
        # vector of elements of its transpose by a permutation, P.
        P = matrix.sqr(
            (
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
            )
        )

        # We can use P to replace var_cov with the covariance matrix of
        # elements of the transpose of B.
        var_cov = P * self._cov_B * P

        # From B = (O^-1)^T we can convert this
        # to the covariance matrix of the real space orthogonalisation matrix
        Bt = self._B.transpose()
        O = Bt.inverse()  # noqa: E741, is name of the matrix
        cov_O = matrix_inverse_error_propagation(Bt, var_cov)

        # The real space unit cell vectors are given by
        vec_a = O * matrix.col((1, 0, 0))
        vec_b = O * matrix.col((0, 1, 0))
        vec_c = O * matrix.col((0, 0, 1))

        # So the unit cell parameters are
        a, b, c = vec_a.length(), vec_b.length(), vec_c.length()
        # alpha = acos(vec_b.dot(vec_c) / (b * c))
        # beta = acos(vec_a.dot(vec_c) / (a * c))
        # gamma = acos(vec_a.dot(vec_b) / (a * b))

        # The estimated errors are calculated by error propagation from cov_O. In
        # each case we define a function F(O) that converts the matrix O into the
        # unit cell parameter of interest. To do error propagation to get the
        # variance of that cell parameter we need the Jacobian of the function.
        # This is a 1*9 matrix of derivatives in the order of the elements of O
        #
        #     / dF   dF   dF   dF   dF   dF   dF   dF   dF  \
        # J = | ---, ---, ---, ---, ---, ---, ---, ---, --- |
        #     \ da1  db1  dc1  da2  db2  dc2  da3  db3  dc3 /
        #

        # For cell length |a|, F = sqrt(a1^2 + a2^2 + a3^2)
        jacobian = matrix.rec(
            (vec_a[0] / a, 0, 0, vec_a[1] / a, 0, 0, vec_a[2] / a, 0, 0), (1, 9)
        )
        var_a = (jacobian * cov_O * jacobian.transpose())[0]

        # For cell length |b|, F = sqrt(b1^2 + b2^2 + b3^2)
        jacobian = matrix.rec(
            (0, vec_b[0] / b, 0, 0, vec_b[1] / b, 0, 0, vec_b[2] / b, 0), (1, 9)
        )
        var_b = (jacobian * cov_O * jacobian.transpose())[0]

        # For cell length |c|, F = sqrt(c1^2 + c2^2 + c3^2)
        jacobian = matrix.rec(
            (0, 0, vec_c[0] / c, 0, 0, vec_c[1] / c, 0, 0, vec_c[2] / c), (1, 9)
        )
        var_c = (jacobian * cov_O * jacobian.transpose())[0]

        # For cell volume (a X b).c,
        # F = c1(a2*b3 - b2*a3) + c2(a3*b1 - b3*a1) + c3(a1*b2 - b1*a2)
        a1, a2, a3 = vec_a
        b1, b2, b3 = vec_b
        c1, c2, c3 = vec_c
        jacobian = matrix.rec(
            (
                c3 * b2 - c2 * b3,
                c1 * b3 - c3 * b1,
                c2 * b1 - c1 * b2,
                c2 * a3 - c3 * a2,
                c3 * a1 - c1 * a3,
                c1 * a2 - c2 * a1,
                a2 * b3 - b2 * a3,
                a3 * b1 - b3 * a1,
                a1 * b2 - b1 * a2,
            ),
            (1, 9),
        )
        var_V = (jacobian * cov_O * jacobian.transpose())[0]
        self._cell_volume_sd = sqrt(var_V)

        # For the unit cell angles we need to calculate derivatives of the angles
        # with respect to the elements of O
        dalpha_db, dalpha_dc = angle_derivative_wrt_vectors(vec_b, vec_c)
        dbeta_da, dbeta_dc = angle_derivative_wrt_vectors(vec_a, vec_c)
        dgamma_da, dgamma_db = angle_derivative_wrt_vectors(vec_a, vec_b)

        # For angle alpha, F = acos( b.c / |b||c|)
        jacobian = matrix.rec(
            (
                0,
                dalpha_db[0],
                dalpha_dc[0],
                0,
                dalpha_db[1],
                dalpha_dc[1],
                0,
                dalpha_db[2],
                dalpha_dc[2],
            ),
            (1, 9),
        )
        var_alpha = (jacobian * cov_O * jacobian.transpose())[0]

        # For angle beta, F = acos( a.c / |a||c|)
        jacobian = matrix.rec(
            (
                dbeta_da[0],
                0,
                dbeta_dc[0],
                dbeta_da[1],
                0,
                dbeta_dc[1],
                dbeta_da[2],
                0,
                dbeta_dc[2],
            ),
            (1, 9),
        )
        var_beta = (jacobian * cov_O * jacobian.transpose())[0]

        # For angle gamma, F = acos( a.b / |a||b|)
        jacobian = matrix.rec(
            (
                dgamma_da[0],
                dgamma_db[0],
                0,
                dgamma_da[1],
                dgamma_db[1],
                0,
                dgamma_da[2],
                dgamma_db[2],
                0,
            ),
            (1, 9),
        )
        var_gamma = (jacobian * cov_O * jacobian.transpose())[0]

        # Symmetry constraints may mean variances of the angles should be zero.
        # Floating point error could lead to negative variances. Ensure these are
        # caught before taking their square root!
        var_alpha = max(0, var_alpha)
        var_beta = max(0, var_beta)
        var_gamma = max(0, var_gamma)

        rad2deg = 180.0 / pi
        self._cell_sd = (
            sqrt(var_a),
            sqrt(var_b),
            sqrt(var_c),
            sqrt(var_alpha) * rad2deg,
            sqrt(var_beta) * rad2deg,
            sqrt(var_gamma) * rad2deg,
        )

    def get_cell_parameter_sd(self):
        """Return the estimated standard deviations of unit cell parameters in
        units of Angstroms and degrees, if available, otherwise None.

        :rtype: :py:class:`tuple`
        """
        if self._cov_B is None:
            return None
        if self._cell_sd is not None:
            return self._cell_sd

        # propagate covariance of B to errors in the unit cell parameters
        self._calc_cell_parameter_sd()

        return self._cell_sd

    def get_cell_volume_sd(self):
        """Return the estimated standard deviation of the unit cell volume in
        units of Angstroms^3, if available, otherwise None.

        :rtype: :py:class:`tuple`
        """
        if self._cov_B is None:
            return None
        if self._cell_volume_sd is not None:
            return self._cell_volume_sd

        # propagate covariance of B to errors in the unit cell parameters and
        # volume
        self._calc_cell_parameter_sd()

        return self._cell_volume_sd

    def set_A(self, A):
        uc = unit_cell(orthogonalization_matrix=A.transpose().inverse())
        B = matrix.sqr(uc.fractionalization_matrix()).transpose()
        U = A * B.inverse()
        self.set_B(B)
        self.set_U(U)

    def set_A_at_scan_points(self, A_list):
        """Set the setting matrix A at a series of checkpoints within a rotation
        scan. There would typically be n+1 points, where n is the number of images
        in the scan. The first point is the state at the beginning of the rotation
        scan. The final point is the state at the end of the rotation scan.
        Intervening points give the state at the start of the rotation at the 2nd
        to the nth image.

        This data is held separately from the 'static' U and B because per-image
        setting matrices may be refined whilst restraining to a previously
        determined best-fit static UB. The values will be reset if any changes are
        made to the static U and B matrices.
        """

        self._A_at_scan_points = [matrix.sqr(e) for e in A_list]
        self._num_scan_points = len(A_list)

    def get_A_at_scan_point(self, t):
        """Return the setting matrix with index t.

        This will typically have been
        set with reference to a particular scan, such that it equals the UB matrix
        appropriate at the start of the rotation for the image with array index t

        :param t: index into the list of scan points.

        :returns: The A matrix at index t
        :rtype: :py:class:`scitbx.matrix.sqr`
        """

        return self._A_at_scan_points[t]

    def get_U_at_scan_point(self, t):
        """
        Return orientation matrix with index t.

        :param t: index into the list of scan points.

        :returns: The U matrix at index t
        :rtype: :py:class:`scitbx.matrix.sqr`
        """

        Bt = self.get_B_at_scan_point(t)
        At = self._A_at_scan_points[t]

        return At * Bt.inverse()

    def get_B_at_scan_point(self, t):
        """
        Return orthogonalisation matrix with index t.

        :param t: index into the list of scan points.

        :returns: The B matrix at index t
        :rtype: :py:class:`scitbx.matrix.sqr`
        """

        At = self._A_at_scan_points[t]
        uc = unit_cell(orthogonalization_matrix=At.transpose().inverse())

        return matrix.sqr(uc.fractionalization_matrix()).transpose()

    def reset_scan_points(self):
        self._num_scan_points = 0
        self._A_at_scan_points = None

    def reset_unit_cell_errors(self):
        self._cov_B = None
        self._cell_sd = None
        self._cell_volume_sd = None

    def get_unit_cell(self):
        """
        :returns: The unit cell
        :rtype: :py:class:`cctbx.uctbx.unit_cell`
        """
        return self._uc

    def set_space_group(self, space_group):
        """
        :param space_group:
        :type space_group: cctbx.sgtbx.space_group
        """
        self._sg = space_group

    def get_space_group(self):
        """
        :returns: The current space group
        :rtype: :py:class:`cctbx.sgtbx.space_group`
        """
        return self._sg

    def get_mosaicity(self, deg=True):
        """
        Get the current value of the mosaicity.

        :param deg: If True, return the mosaicity in degrees, else return the
                    mosaicity in radians.
        :type deg: bool
        :returns: The mosaicity
        :rtype: float or None
        """
        if deg is True:
            return self._mosaicity * 180.0 / pi

        return self._mosaicity

    def set_mosaicity(self, mosaicity, deg=True):
        """
        Update the value of the mosaicity.

        :param mosaicity: The mosaicity
        :type mosaicity: float
        :param deg: If True, assume the mosaicity is given in degrees, else
                    assume it is given in radians.
        :type deg: bool
        :returns: The mosaicity
        :rtype: float or None
        """
        if deg is True:
            self._mosaicity = mosaicity * pi / 180.0
        else:
            self._mosaicity = mosaicity

    def get_A(self):
        return self._U * self._B

    def __eq__(self, other, eps=1e-7):
        raise NotImplementedError("this piece of code was broken")

    def __ne__(self, other):
        return not self.__eq__(other)

    __hash__ = None

    def is_similar_to(
        self,
        other,
        angle_tolerance=0.01,
        mosaicity_tolerance=0.8,
        uc_rel_length_tolerance=0.01,
        uc_abs_angle_tolerance=1.0,
    ):
        """
        Test similarity of this to another crystal model

        :param other: the crystal model to test against
        :type other: crystal_model
        :param angle_tolerance: maximum tolerated orientation difference in degrees
        :type angle_tolerance: float
        :param mosaicity_tolerance: minimum tolerated fraction of the larger
                mosaicity for the smaller mosaicity to retain similarity
        :type mosaicity_tolerance:float
        :param uc_rel_length_tolerance: relative length tolerance to pass to
                unit_cell.is_similar_to
        :type uc_rel_length_tolerance: float
        :param uc_abs_angle_tolerance: absolute angle tolerance to pass to
                unit_cell.is_similar_to
        :type uc_abs_angle_tolerance: float
        :returns: Whether the other crystal model is similar to this
        :rtype: bool
        """

        # space group test
        if self.get_space_group() != other.get_space_group():
            return False

        # mosaicity test
        m_a, m_b = self.get_mosaicity(deg=True), other.get_mosaicity(deg=True)
        min_m, max_m = min(m_a, m_b), max(m_a, m_b)
        if min_m <= 0.0:
            if max_m > 0.0:
                return False
        elif min_m / max_m < mosaicity_tolerance:
            return False

        # static orientation test
        U_a, U_b = self.get_U(), other.get_U()
        assert U_a.is_r3_rotation_matrix()
        assert U_b.is_r3_rotation_matrix()
        R_ab = U_b * U_a.transpose()
        uq = R_ab.r3_rotation_matrix_as_unit_quaternion()
        angle = uq.unit_quaternion_as_axis_and_angle(deg=True)[0]
        if abs(angle) > angle_tolerance:
            return False

        # static unit cell test
        uc_a, uc_b = self.get_unit_cell(), other.get_unit_cell()
        if not uc_a.is_similar_to(
            uc_b,
            relative_length_tolerance=uc_rel_length_tolerance,
            absolute_angle_tolerance=uc_abs_angle_tolerance,
        ):
            return False

        # scan varying tests
        if self.num_scan_points > 0:
            if other.num_scan_points != self.num_scan_points:
                return False
            for i in range(self.num_scan_points):
                U_a, U_b = self.get_U_at_scan_point(i), other.get_U_at_scan_point(i)
                assert U_a.is_r3_rotation_matrix()
                assert U_b.is_r3_rotation_matrix()
                R_ab = U_b * U_a.transpose()
                uq = R_ab.r3_rotation_matrix_as_unit_quaternion()
                angle = uq.unit_quaternion_as_axis_and_angle(deg=True)[0]
                if abs(angle) > angle_tolerance:
                    return False

                B_a, B_b = self.get_B_at_scan_point(i), other.get_B_at_scan_point(i)
                uc_a = crystal_orientation(B_a, True).unit_cell()
                uc_b = crystal_orientation(B_b, True).unit_cell()
                if not uc_a.is_similar_to(
                    uc_b,
                    relative_length_tolerance=uc_rel_length_tolerance,
                    absolute_angle_tolerance=uc_abs_angle_tolerance,
                ):
                    return False

        return True

    def get_real_space_vectors(self):
        """
        Get the real space unit cell basis vectors.

        :returns: Real space unit cell basis vectors
        :rtype: list of :py:class:`scitbx.matrix.col`
        """
        A_inv = self.get_A().inverse()
        return (matrix.col(A_inv[:3]), matrix.col(A_inv[3:6]), matrix.col(A_inv[6:9]))

    def change_basis(self, change_of_basis_op):
        raise NotImplementedError("this piece of code was broken")

    def update(self, other):
        """
        Update the current crystal model such that self == other.

        :param other: A crystal model to update self with.
        :type other: crystal_model
        """
        self._U = other._U
        self._B = other._B
        self._A_at_scan_points = other._A_at_scan_points
        self._num_scan_points = other._num_scan_points
        self._sg = other._sg
        self._uc = other._uc
        self._mosaicity = other._mosaicity

    def rotate_around_origin(self, axis, angle, deg=True):
        """
        Rotate the model around an axis and angle

        :param axis: The axis to rotate around
        :param angle: The angle to rotate around
        :param deg: Degrees or radians

        """
        # Compute the matrix
        R = matrix.col(axis).axis_and_angle_as_r3_rotation_matrix(angle, deg=deg)

        # Update U
        self._U = R * self._U

        # Update A at scan points
        if self._A_at_scan_points is not None:
            for i, At in enumerate(self._A_at_scan_points):
                uc = unit_cell(orthogonalization_matrix=At.transpose().inverse())
                Bt = matrix.sqr(uc.fractionalization_matrix()).transpose()
                Ut = At * Bt.inverse()
                Ut = R * Ut
                At = Ut * Bt
                self._A_at_scan_points[i] = At



from __future__ import annotations

import h5py
import hdf5plugin
import numpy
import pytest

from scitbx.array_family import flex  # noqa: F401; boost python bindings

from dxtbx_format_nexus_ext import (
    dataset_as_flex_double,
    dataset_as_flex_float,
    dataset_as_flex_int,
)


def uncompressed(file, shape, type_name):
    return file.create_dataset(
        "data",
        shape,
        dtype=type_name,
    )


def gzip(file, shape, type_name):
    return file.create_dataset(
        "data",
        shape,
        compression="gzip",
        dtype=type_name,
    )


def bshuf_lz4(file, shape, type_name):
    return file.create_dataset(
        "data",
        shape,
        **hdf5plugin.Bitshuffle(),
        dtype=type_name,
    )


@pytest.mark.parametrize(
    "creator",
    [
        uncompressed,
        gzip,
        bshuf_lz4,
    ],
)
@pytest.mark.parametrize(
    "type_name,converter",
    [
        ("int", dataset_as_flex_int),
        ("float", dataset_as_flex_float),
        ("double", dataset_as_flex_double),
    ],
)
def test_dataset_as_flex(type_name, creator, converter):
    # Create an in-memory HDF5 dataset with unique name
    f = h5py.File(type_name + ".h5", "w", driver="core", backing_store=False)

    shape = (20, 20, 20)
    dataset = creator(f, shape, type_name)

    # create some random data
    array = numpy.random.rand(*shape) * 100
    original = array.astype(type_name)

    dataset[:] = original

    selection = slice(0, shape[0], 1), slice(0, shape[1], 1), slice(0, shape[2], 1)
    foo = converter(dataset.id.id, selection)

    assert foo.as_numpy_array() == pytest.approx(original)



from __future__ import annotations

import math
import os
import pickle

import pytest

from scitbx import matrix
from scitbx.array_family import flex
from scitbx.math import principal_axes_of_inertia_2d

from dxtbx.masking import (
    GoniometerMaskerFactory,
    is_inside_polygon,
    mask_untrusted_polygon,
)
from dxtbx.model.detector import DetectorFactory
from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx.model.goniometer import GoniometerFactory


def test_polygon():
    poly = flex.vec2_double(((0, 0), (1, 0), (1, 1), (0, 1)))
    assert is_inside_polygon(poly, 0, 0)
    assert is_inside_polygon(poly, 0.5, 0.5)
    assert not is_inside_polygon(poly, 1, 1.01)
    points = flex.vec2_double(((0.3, 0.8), (0.3, 1.5), (-8, 9), (0.00001, 0.9999)))
    assert list(is_inside_polygon(poly, points)) == [True, False, False, True]


@pytest.fixture
def kappa_goniometer():
    def _construct_goniometer(phi, kappa, omega):
        phi_axis = (1.0, 0.0, 0.0)
        kappa_axis = (0.914, 0.279, -0.297)
        omega_axis = (1.0, 0.0, 0.0)
        axes = flex.vec3_double((phi_axis, kappa_axis, omega_axis))
        angles = flex.double((phi, kappa, omega))
        names = flex.std_string(("GON_PHI", "GON_KAPPA", "GON_OMEGA"))
        return GoniometerFactory.make_multi_axis_goniometer(
            axes, angles, names, scan_axis=2
        )

    return _construct_goniometer


@pytest.fixture
def smargon_goniometer():
    def _construct_goniometer(phi, chi, omega):
        phi_axis = (1.0, 0.0, 0.0)
        chi_axis = (0, 0, -1)
        omega_axis = (1.0, 0.0, 0.0)
        axes = flex.vec3_double((phi_axis, chi_axis, omega_axis))
        angles = flex.double((phi, chi, omega))
        names = flex.std_string(("GON_PHI", "GON_CHI", "GON_OMEGA"))
        return GoniometerFactory.make_multi_axis_goniometer(
            axes, angles, names, scan_axis=2
        )

    return _construct_goniometer


@pytest.fixture
def pilatus_6M():
    def _construct_detector(distance):
        return DetectorFactory.simple(
            sensor="PAD",
            distance=distance,
            beam_centre=(216.87, 211.32),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(0.172, 0.172),
            image_size=(2463, 2527),
            trusted_range=(-1, 1e8),
        )

    return _construct_detector


@pytest.fixture(params=["cpp", "python"])
def kappa_goniometer_shadow_masker(request):
    def _construct_shadow_masker(goniometer):
        if request.param == "cpp":
            return GoniometerMaskerFactory.mini_kappa(goniometer)

        # Simple model of cone around goniometer phi axis
        # Exact values don't matter, only the ratio of height/radius
        height = 50  # mm
        radius = 20  # mm

        steps_per_degree = 1
        theta = (
            flex.double(range(360 * steps_per_degree))
            * math.pi
            / 180
            * 1
            / steps_per_degree
        )
        y = radius * flex.cos(-theta)
        z = radius * flex.sin(-theta)
        x = flex.double(theta.size(), height)

        coords = flex.vec3_double(zip(x, y, z))
        coords.insert(0, (0, 0, 0))

        return PyGoniometerShadowMasker(goniometer, coords, flex.size_t(len(coords), 0))

    return _construct_shadow_masker


@pytest.fixture(params=["cpp", "python"])
def smargon_shadow_masker(request):
    def _construct_shadow_masker(goniometer):
        if request.param == "python":
            return PySmarGonShadowMasker(goniometer)
        return GoniometerMaskerFactory.smargon(goniometer)

    return _construct_shadow_masker


@pytest.fixture
def dls_i23_experiment(dials_regression):
    experiments_file = os.path.join(
        dials_regression, "shadow_test_data/DLS_I23_Kappa/data_1_0400.cbf.gz"
    )
    experiments = ExperimentListFactory.from_filenames([experiments_file])
    return experiments[0]


@pytest.fixture
def dls_i23_kappa_shadow_masker():
    def _construct_shadow_masker(goniometer):
        return GoniometerMaskerFactory.dls_i23_kappa(goniometer)

    return _construct_shadow_masker


@pytest.fixture
def dls_i19_2_detector():
    def _construct_detector(distance):
        return DetectorFactory.simple(
            sensor="PAD",
            distance=distance,
            beam_centre=(41.20, 51.69),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(0.172, 0.172),
            image_size=(487, 619),
            trusted_range=(-1, 1e8),
        )

    return _construct_detector


@pytest.fixture
def dls_i19_2_goniometer():
    def _construct_goniometer(phi, kappa, omega):
        phi_axis = (1.0, 0.0, 0.0)
        kappa_axis = (0.642788, -0.766044, 0)
        omega_axis = (1.0, 0.0, 0.0)
        axes = flex.vec3_double((phi_axis, kappa_axis, omega_axis))
        angles = flex.double((phi, kappa, omega))
        names = flex.std_string(("GON_PHI", "GON_KAPPA", "GON_OMEGA"))
        return GoniometerFactory.make_multi_axis_goniometer(
            axes, angles, names, scan_axis=2
        )

    return _construct_goniometer


@pytest.fixture
def diamond_anvil_cell_masker():
    def _construct_shadow_masker(goniometer):
        return GoniometerMaskerFactory.diamond_anvil_cell(
            goniometer, cone_opening_angle=2 * 38 * math.pi / 180
        )

    return _construct_shadow_masker


def test_GoniometerShadowMasker_kappa_180_omega_0(
    kappa_goniometer, pilatus_6M, kappa_goniometer_shadow_masker
):
    goniometer = kappa_goniometer(phi=0, kappa=180, omega=0)
    detector = pilatus_6M(distance=170)
    masker = kappa_goniometer_shadow_masker(goniometer)

    scan_angle = 0  # omega = 0, shadow in top right corner of detector
    extrema = masker.extrema_at_scan_angle(scan_angle)
    assert len(extrema) == 361
    assert extrema[1] == pytest.approx(
        (43.60448791048147, 8.572923552543028, -30.416337975287743)
    )

    shadow = masker.project_extrema(detector, scan_angle)
    assert len(shadow) == len(detector)
    assert len(shadow[0]) == 145

    mask = masker.get_mask(detector, scan_angle)
    assert len(mask) == len(detector)
    assert mask[0].all() == tuple(reversed(detector[0].get_image_size()))
    assert mask[0].count(True) == pytest.approx(5570865)

    obj = pickle.dumps(masker)
    masker2 = pickle.loads(obj)

    mask2 = masker2.get_mask(detector, scan_angle)
    assert len(mask2) == len(mask)
    assert mask2[0].all() == mask[0].all()
    assert mask2[0].count(True) == mask2[0].count(True)


def test_GoniometerShadowMasker_kappa_180_omega_m45(
    kappa_goniometer, pilatus_6M, kappa_goniometer_shadow_masker
):
    goniometer = kappa_goniometer(phi=0, kappa=180, omega=0)
    detector = pilatus_6M(distance=170)
    masker = kappa_goniometer_shadow_masker(goniometer)

    scan_angle = -45  # omega = -45, shadow on centre-right of detector
    extrema = masker.extrema_at_scan_angle(scan_angle)
    assert len(extrema) == 361
    assert extrema[1] == pytest.approx(
        (43.60448791048147, -15.445626462590827, -27.569571219784905)
    )
    shadow = masker.project_extrema(detector, scan_angle)
    assert len(shadow[0]) in (226, 227)
    # assert shadow[0][0] == pytest.approx((1623.5133257425446, 1254.0966982780835))
    mask = masker.get_mask(detector, scan_angle)
    assert mask[0].all() == tuple(reversed(detector[0].get_image_size()))
    assert mask[0].count(True) == pytest.approx(5467810)

    obj = pickle.dumps(masker)
    masker2 = pickle.loads(obj)

    mask2 = masker2.get_mask(detector, scan_angle)
    assert len(mask2) == len(mask)
    assert mask2[0].all() == mask[0].all()
    assert mask2[0].count(True) == mask2[0].count(True)


def test_GoniometerShadowMasker_kappa_180_omega_p45(
    kappa_goniometer, pilatus_6M, kappa_goniometer_shadow_masker
):
    goniometer = kappa_goniometer(phi=0, kappa=180, omega=0)
    detector = pilatus_6M(distance=170)
    masker = kappa_goniometer_shadow_masker(goniometer)

    scan_angle = 45  # omega = +45, no shadow on detector
    shadow = masker.project_extrema(detector, scan_angle)
    assert len(shadow[0]) == 0
    extrema = masker.extrema_at_scan_angle(scan_angle)
    assert len(extrema) == 361
    assert extrema[1] == pytest.approx(
        (43.60448791048147, 27.56957121978491, -15.445626462590822)
    )
    mask = masker.get_mask(detector, scan_angle)
    assert mask[0] is None or mask[0].count(False) == 0

    obj = pickle.dumps(masker)
    masker2 = pickle.loads(obj)

    mask2 = masker2.get_mask(detector, scan_angle)
    assert mask2[0] is None or mask2[0].count(False) == 0


def test_GoniometerShadowMasker_kappa_m70_omega_p100(
    kappa_goniometer, pilatus_6M, kappa_goniometer_shadow_masker
):
    # goniometer shadow does not intersect with detector panel
    goniometer = kappa_goniometer(phi=0, kappa=-70, omega=0)
    detector = pilatus_6M(distance=170)
    masker = kappa_goniometer_shadow_masker(goniometer)

    scan_angle = 100
    extrema = masker.extrema_at_scan_angle(scan_angle)
    assert len(extrema) == 361
    assert extrema[1] == pytest.approx(
        (42.318198019878935, 8.61724085750561, 32.17006801910824)
    )
    shadow = masker.project_extrema(detector, scan_angle)
    assert len(shadow[0]) == 0
    mask = masker.get_mask(detector, scan_angle)
    assert mask[0] is None or mask[0].count(False) == 0

    obj = pickle.dumps(masker)
    masker2 = pickle.loads(obj)

    mask2 = masker2.get_mask(detector, scan_angle)
    assert mask2[0] is None or mask2[0].count(False) == 0


def test_SmarGonShadowMasker_p48_c45_o95(
    smargon_goniometer, pilatus_6M, smargon_shadow_masker
):
    goniometer = smargon_goniometer(phi=48, chi=45, omega=100)
    detector = pilatus_6M(distance=170)
    masker = smargon_shadow_masker(goniometer)

    scan_angle = 100
    extrema = masker.extrema_at_scan_angle(scan_angle)
    assert len(extrema) == 82
    assert extrema[1] == pytest.approx(
        (22.106645739466337, 13.963679418895005, -22.47914302717216)
    )
    shadow = masker.project_extrema(detector, scan_angle)
    assert len(shadow[0]) in (15, 16)
    mask = masker.get_mask(detector, scan_angle)
    assert mask[0].count(True) == pytest.approx(5716721, 3e-5)

    obj = pickle.dumps(masker)
    masker2 = pickle.loads(obj)

    mask2 = masker2.get_mask(detector, scan_angle)
    assert len(mask2) == len(mask)
    assert mask2[0].all() == mask[0].all()
    assert mask2[0].count(True) == mask2[0].count(True)


def test_SmarGonShadowMasker_p0_c90_o50(
    smargon_goniometer, pilatus_6M, smargon_shadow_masker
):
    for phi in (0, 90, 180):
        # There should be no dependence of the shadow on phi
        goniometer = smargon_goniometer(phi=phi, chi=90, omega=50)
        detector = pilatus_6M(distance=214.75)
        masker = smargon_shadow_masker(goniometer)

        scan_angle = 50
        extrema = masker.extrema_at_scan_angle(scan_angle)
        assert len(extrema) == 82
        assert extrema[1] == pytest.approx(
            (-1.7364817766692957, -13.66792605230091, -31.609688838521162)
        )
        shadow = masker.project_extrema(detector, scan_angle)
        assert len(shadow[0]) == 11
        mask = masker.get_mask(detector, scan_angle)
        assert mask[0].count(True) == pytest.approx(4614865, 2e-4)

        obj = pickle.dumps(masker)
        masker2 = pickle.loads(obj)

        mask2 = masker2.get_mask(detector, scan_angle)
        assert len(mask2) == len(mask)
        assert mask2[0].all() == mask[0].all()
        assert mask2[0].count(True) == mask2[0].count(True)


def test_dls_i23_kappa(dls_i23_experiment, dls_i23_kappa_shadow_masker):
    for phi in (0, 90, 180):
        goniometer = dls_i23_experiment.goniometer
        goniometer.set_angles((0, -180, 0))
        detector = dls_i23_experiment.detector
        masker = dls_i23_kappa_shadow_masker(goniometer)

        scan_angle = 40
        extrema = masker.extrema_at_scan_angle(scan_angle)
        assert len(extrema) == 66
        assert extrema[1] == pytest.approx(
            (-18.68628039873836, -55.47017980234442, -98.76129898677576)
        )
        shadow = masker.project_extrema(detector, scan_angle)
        assert len(shadow) == len(detector)
        assert [len(s) for s in shadow] == [
            0,
            0,
            0,
            6,
            9,
            7,
            7,
            7,
            9,
            4,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ]
        mask = masker.get_mask(detector, scan_angle)
        assert sum(m.count(False) for m in mask) == 1898843

        obj = pickle.dumps(masker)
        masker2 = pickle.loads(obj)

        mask2 = masker2.get_mask(detector, scan_angle)
        assert len(mask2) == len(mask)
        assert mask2[0].all() == mask[0].all()
        assert mask2[0].count(True) == mask2[0].count(True)


def test_dls_i19_2_diamond_anvil_cell(
    dls_i19_2_detector, dls_i19_2_goniometer, diamond_anvil_cell_masker
):
    detector = dls_i19_2_detector(distance=90.29)
    goniometer = dls_i19_2_goniometer(-35, 0, -90)
    masker = diamond_anvil_cell_masker(goniometer)
    scan_angle = -35

    extrema = masker.extrema_at_scan_angle(scan_angle)
    assert len(extrema) == 721
    assert extrema[1] == pytest.approx(
        (7.812856265067172, 3.4202014332566875, -9.396926207859082)
    )
    shadow = masker.project_extrema(detector, scan_angle)
    assert len(shadow) == len(detector)
    assert len(shadow[0]) == 90
    mask = masker.get_mask(detector, scan_angle)
    assert sum(m.count(False) for m in mask) == 84352

    obj = pickle.dumps(masker)
    masker2 = pickle.loads(obj)

    mask2 = masker2.get_mask(detector, scan_angle)
    assert len(mask2) == len(mask)
    assert mask2[0].all() == mask[0].all()
    assert mask2[0].count(True) == mask2[0].count(True)


class PyGoniometerShadowMasker:
    def __init__(self, goniometer, extrema_at_datum, axis):
        # axis is an array of size_t the same size as extrema_at_datum,
        # where each element identifies the axis that that coordinate depends on
        self.goniometer = goniometer
        self._extrema_at_datum = extrema_at_datum
        self.axis = axis

    def extrema_at_scan_angle(self, scan_angle):
        axes = self.goniometer.get_axes()
        angles = self.goniometer.get_angles()
        scan_axis = self.goniometer.get_scan_axis()
        angles[scan_axis] = scan_angle
        extrema = self._extrema_at_datum.deep_copy()

        for i in range(len(axes)):
            sel = self.axis <= i
            rotation = matrix.col(axes[i]).axis_and_angle_as_r3_rotation_matrix(
                angles[i], deg=True
            )
            extrema.set_selected(sel, rotation.elems * extrema.select(sel))

        return extrema

    def project_extrema(self, detector, scan_angle):
        coords = self.extrema_at_scan_angle(scan_angle)
        shadow_boundary = []

        for p_id, p in enumerate(detector):
            # project coordinates onto panel plane
            a = p.get_D_matrix() * coords
            x, y, z = a.parts()
            valid = z > 0
            x.set_selected(valid, x.select(valid) / z.select(valid))
            y.set_selected(valid, y.select(valid) / z.select(valid))

            if valid.count(True) < 3:
                # no shadow projected onto this panel
                shadow_boundary.append(flex.vec2_double())
                continue

            # Compute convex hull of shadow points
            points = flex.vec2_double(x.select(valid), y.select(valid))
            shadow = flex.vec2_double(_convex_hull(points))
            shadow *= 1 / p.get_pixel_size()[0]

            shadow_orig = shadow.deep_copy()

            for i in (0, p.get_image_size()[0]):
                points = flex.vec2_double(
                    flex.double(p.get_image_size()[1], i),
                    flex.double_range(0, p.get_image_size()[1]),
                )
                inside = is_inside_polygon(shadow_orig, points)
                # only add those points needed to define vertices of shadow
                inside_isel = inside.iselection()
                outside_isel = (~inside).iselection()
                while inside_isel.size():
                    j = inside_isel[0]
                    shadow.append(points[j])
                    outside_isel = outside_isel.select(outside_isel > j)
                    if outside_isel.size() == 0:
                        shadow.append(points[inside_isel[-1]])
                        break
                    sel = inside_isel >= outside_isel[0]
                    if sel.count(True) == 0:
                        shadow.append(points[inside_isel[-1]])
                        break
                    inside_isel = inside_isel.select(sel)

            for i in (0, p.get_image_size()[1]):
                points = flex.vec2_double(
                    flex.double_range(0, p.get_image_size()[0]),
                    flex.double(p.get_image_size()[0], i),
                )
                inside = is_inside_polygon(shadow_orig, points)
                # only add those points needed to define vertices of shadow
                inside_isel = inside.iselection()
                outside_isel = (~inside).iselection()
                while inside_isel.size():
                    j = inside_isel[0]
                    shadow.append(points[j])
                    outside_isel = outside_isel.select(outside_isel > j)
                    if outside_isel.size() == 0:
                        shadow.append(points[inside_isel[-1]])
                        break
                    sel = inside_isel >= outside_isel[0]
                    if sel.count(True) == 0:
                        shadow.append(points[inside_isel[-1]])
                        break
                    inside_isel = inside_isel.select(sel)

            # Select only those vertices that are within the panel dimensions
            n_px = p.get_image_size()
            x, y = shadow.parts()
            valid = (x >= 0) & (x <= n_px[0]) & (y >= 0) & (y <= n_px[1])
            shadow = shadow.select(valid)

            # sort vertices clockwise from centre of mass
            com = principal_axes_of_inertia_2d(shadow).center_of_mass()
            sx, sy = shadow.parts()
            shadow = shadow.select(
                flex.sort_permutation(flex.atan2(sy - com[1], sx - com[0]))
            )

            shadow_boundary.append(shadow)

        return shadow_boundary

    def get_mask(self, detector, scan_angle):
        shadow_boundary = self.project_extrema(detector, scan_angle)

        mask = []
        for panel_id, panel in enumerate(detector):
            m = None
            if shadow_boundary[panel_id].size() > 3:
                m = flex.bool(flex.grid(reversed(panel.get_image_size())), True)
                mask_untrusted_polygon(m, shadow_boundary[panel_id])
            mask.append(m)
        return mask


# https://en.wikibooks.org/wiki/Algorithm_Implementation/Geometry/Convex_hull/Monotone_chain#Python
# https://github.com/thepracticaldev/orly-full-res/blob/master/copyingandpasting-big.png
def _convex_hull(points):
    """Computes the convex hull of a set of 2D points.

    Input: an iterable sequence of (x, y) pairs representing the points.
    Output: a list of vertices of the convex hull in counter-clockwise order,
      starting from the vertex with the lexicographically smallest coordinates.
    Implements Andrew's monotone chain algorithm. O(n log n) complexity.
    """

    # Sort the points lexicographically (tuples are compared lexicographically).
    # Remove duplicates to detect the case we have just one unique point.
    points = sorted(set(points))

    # Boring case: no points or a single point, possibly repeated multiple times.
    if len(points) <= 1:
        return points

    # 2D cross product of OA and OB vectors, i.e. z-component of their 3D cross product.
    # Returns a positive value, if OAB makes a counter-clockwise turn,
    # negative for clockwise turn, and zero if the points are collinear.
    def cross(o, a, b):
        return (a[0] - o[0]) * (b[1] - o[1]) - (a[1] - o[1]) * (b[0] - o[0])

    # Build lower hull
    lower = []
    for p in points:
        while len(lower) >= 2 and cross(lower[-2], lower[-1], p) <= 0:
            lower.pop()
        lower.append(p)

    # Build upper hull
    upper = []
    for p in reversed(points):
        while len(upper) >= 2 and cross(upper[-2], upper[-1], p) <= 0:
            upper.pop()
        upper.append(p)

    # Concatenation of the lower and upper hulls gives the convex hull.
    # Last point of each list is omitted because it is repeated at the beginning of the other list.
    return lower[:-1] + upper[:-1]


class PySmarGonShadowMasker(PyGoniometerShadowMasker):
    def __init__(self, goniometer):
        # FACE A: Sample holder
        #   Defined as semi-circle of radius r(A) = 10 mm (centred on PHI axis)
        #   with rectangle of size a(A) = 12.8 mm (x 20 mm)

        offsetA = 33.0
        # semi-circle for phi=-90 ... +90
        radiusA = 10.0
        phi = flex.double_range(-90, 100, step=10) * math.pi / 180
        x = flex.double(phi.size(), -offsetA)
        y = radiusA * flex.cos(phi)
        z = radiusA * flex.sin(phi)

        # corners of square
        sqdA = 12.8  # square depth
        nsteps = 10
        for i in range(nsteps + 1):
            for sign in (+1, -1):
                x.append(-offsetA)
                y.append(i * -sqdA / nsteps)
                z.append(sign * radiusA)
        x.append(-offsetA)
        y.append(-sqdA)
        z.append(0)

        self.faceA = flex.vec3_double(-x, -y, z)

        # FACE B: Lower arm
        sx = -28.50
        sy = -4.90
        sz = 8.50
        mx = -13.80
        my = -26.00
        nx = -27.50
        ny = -29.50
        px = -65.50
        py = -29.50
        self.faceB = flex.vec3_double(
            ((-sx, -sy, sz), (-mx, -my, 0), (-nx, -ny, 0), (-px, -py, 0))
        )

        # FACE E: Rim of sample holder
        #   Defined as circle of radius r(E) = 6 mm (centred on PHI axis) at an
        #   offset o(E) = 19 mm

        offsetE = 19.0
        radiusE = 6.0
        phi = flex.double_range(0, 360, step=15) * math.pi / 180
        x = flex.double(phi.size(), -offsetE)
        y = radiusE * flex.cos(phi)
        z = radiusE * flex.sin(phi)

        self.faceE = flex.vec3_double(-x, -y, z)

        extrema_at_datum = self.faceA.deep_copy()
        extrema_at_datum.extend(self.faceE)
        super().__init__(
            goniometer, extrema_at_datum, flex.size_t(extrema_at_datum.size(), 1)
        )

    def extrema_at_scan_angle(self, scan_angle):
        extrema = super().extrema_at_scan_angle(scan_angle)

        axes = self.goniometer.get_axes()
        angles = self.goniometer.get_angles()
        scan_axis = self.goniometer.get_scan_axis()
        angles[scan_axis] = scan_angle

        s = matrix.col(self.faceB[0])
        mx, my, _ = self.faceB[1]
        nx, ny, _ = self.faceB[2]
        px, py, _ = self.faceB[3]

        Rchi = matrix.col(axes[1]).axis_and_angle_as_r3_rotation_matrix(
            angles[1], deg=True
        )
        sk = Rchi * s
        sxk, syk, szk = sk.elems
        coords = flex.vec3_double(
            (
                (sxk, syk, 0),
                (sxk, syk, szk),
                (sxk + mx / 2, syk + my / 2, szk),
                (sxk + mx, syk + my, szk),
                (sxk + (mx + nx) / 2, syk + (my + ny) / 2, szk),
                (sxk + nx, syk + ny, szk),
                (sxk + (nx + px) / 2, syk + (ny + py) / 2, szk),
                (sxk + px, syk + py, szk),
                (sxk + px, syk + py, 0),
                (sxk + px, syk + py, -szk),
                (sxk + (nx + px) / 2, syk + (ny + py) / 2, -szk),
                (sxk + nx, syk + ny, -szk),
                (sxk + (mx + nx) / 2, syk + (my + ny) / 2, -szk),
                (sxk + mx, syk + my, -szk),
                (sxk + mx / 2, syk + my / 2, -szk),
                (sxk, syk, -szk),
            )
        )

        Romega = matrix.col(axes[2]).axis_and_angle_as_r3_rotation_matrix(
            angles[2], deg=True
        )
        coords = Romega.elems * coords
        extrema.extend(coords)

        return extrema



from __future__ import annotations

import os

from libtbx.test_utils import approx_equal

from dxtbx.imageset import ImageSetFactory


def test_multi_axis_goniometer(dials_regression):
    data_dir = os.path.join(dials_regression, "image_examples", "dials-190")

    imgset = ImageSetFactory.new(os.path.join(data_dir, "whatev1_01_00001.cbf"))[0]
    gonio = imgset.get_goniometer(0)
    assert approx_equal(gonio.get_fixed_rotation(), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    assert approx_equal(gonio.get_setting_rotation(), (1, 0, 0, 0, 1, 0, 0, 0, 1))

    imgset = ImageSetFactory.new(os.path.join(data_dir, "whatev1_02_00001.cbf"))[0]
    gonio = imgset.get_goniometer(0)
    assert approx_equal(gonio.get_fixed_rotation(), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    assert approx_equal(
        gonio.get_setting_rotation(),
        (1, 0, 0, 0, -0.5, 0.866, 0.0, -0.866, -0.5),
        eps=1e-4,
    )

    imgset = ImageSetFactory.new(os.path.join(data_dir, "whatev1_03_00001.cbf"))[0]
    gonio = imgset.get_goniometer(0)
    assert approx_equal(gonio.get_fixed_rotation(), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    assert approx_equal(
        gonio.get_setting_rotation(),
        (1, 0, 0, 0, 0.5, 0.866, 0.0, -0.866, 0.5),
        eps=1e-4,
    )


def test_still(dials_regression):
    data_dir = os.path.join(dials_regression, "image_examples", "DLS_I04")

    imgset = ImageSetFactory.new(os.path.join(data_dir, "grid_full_cbf_0005.cbf"))[0]
    if imgset.get_scan(0):
        scan = imgset.get_scan(0)
        assert approx_equal(scan.get_image_oscillation(0), (30.0, 0.0), eps=1e-5)
    beam = imgset.get_beam(0)
    beam.get_s0()
    assert approx_equal(beam.get_s0(), (-0.0, -0.0, -1.0209290454313424))



"""
Image reading tests against the dials_regression suite
"""

from __future__ import annotations

import bz2
import gzip
import os
import shutil
from pathlib import Path

import pytest

import scitbx.matrix
from rstbx.slip_viewer.slip_viewer_image_factory import SlipViewerImageFactory
from scitbx.array_family import flex

import dxtbx.format.Registry
from dxtbx.util import get_url_scheme

try:
    import h5py
except ModuleNotFoundError:
    h5py = None

try:
    import cbflib_adaptbx

    # Check this isn't a namespace package
    if hasattr(cbflib_adaptbx, "__path__") and cbflib_adaptbx.__file__ is None:
        cbflib_adaptbx = None
except ModuleNotFoundError:
    # Detectorbase for CBF won't work
    cbflib_adaptbx = None

_files = (
    "ALS_1231/q315r_lyso_1_001.img",
    "ALS_422/lyso_041013a_1_001.img",
    "ALS_501/als501_q4_1_001.img",
    "ALS_733/200mMNaCl5pcGlyc_400.edf",
    "ALS_821/q210_lyso_1_101.img",
    "ALS_831/q315r_lyso_001.img",
    "APS_14BMC/q315_1_001.img",
    "APS_17ID/q210_1_001.img",
    "APS_19ID/q315_unbinned_a.0001.img",
    "APS_19ID/t1.0001.img.bz2",
    "APS_22ID/mar300.0001",
    "APS_23IDD/mar300_1_E1.0001",
    "APS_24IDC/pilatus_1_0001.cbf",
    "APS_24IDC/q315_1_001.img",
    "APS_24IDE_test/thaum-12_1_0001.cbf",
    "APS_24IDE_test/thaum-12_1_0002.cbf",
    "APS_24IDE_test/thaum-12_1_0003.cbf",
    "APS_24IDE_test/thaum-12_1_0004.cbf",
    "APS_24IDE_test/thaum-12_1_0005.cbf",
    "APS_24IDE_test/thaum-12_1_0006.cbf",
    "APS_24IDE_test/thaum-12_1_0007.cbf",
    "APS_24IDE_test/thaum-12_1_0008.cbf",
    "APS_24IDE_test/thaum-12_1_0009.cbf",
    "APS_24IDE_test/thaum-12_1_0010.cbf",
    "Bruker_PHOTON_II/dan_01_0001.sfrm",
    "CLS1_08ID1/mar225_2_E0_0001.img",
    "DESY_BW7B/mar345_01_001.mar2300",
    "DESY_ID141/q210_2_001.img",
    "DLS_I02/X4_wide_M1S4_1_0001.cbf",
    "DLS_I04/grid_full_cbf_0005.cbf",
    "DLS_I19/I19_P300k_00001.cbf",
    "DLS_I23/I23_P12M_alpha_0001.cbf",
    "DLS_I23/germ_13KeV_0001.cbf",
    "DLS_I24_stills/still_0001.cbf",
    "DLS_eBIC/nanB_4_singleframe_1.ser.bz2",
    "DTREK_home_lab/s01f0001.osc",
    "ESRF_BM14/mar165_001.mccd",
    "ESRF_BM14/mar225_1_001.mccd",
    "ESRF_ID231/q315r_7_001.img",
    "ESRF_ID29/trypsin_1_0001.cbf",
    "LCLS_cspad_nexus/idx-20130301060858401.cbf",
    "LCLS_cspad_nexus/idx-20130301060858601.cbf",
    "LCLS_cspad_nexus/idx-20130301060858701.cbf",
    "LCLS_cspad_nexus/idx-20130301060858801.cbf",
    "LCLS_jungfrau/jungfrau_multipanel.cbf",
    "MLFSOM_simulation/fake_00001.img",
    "MacScience/reallysurprise_001.ipf",
    "RAXIS-HTC/test1_lysozyme_0111060001.osc",
    "RigakuA200/XV10001.img",
    "SACLA_MPCCD_Cheetah/run266702-0-subset.h5",
    "SLS_Eiger_16M_as_CBF/insu_with_bs_labelit_0001.cbf",
    "SLS_Eiger_16M_as_CBF/insu_with_bs_labelit_0901.cbf",
    "SLS_X06SA/mar225_2_001.img",
    "SLS_X06SA/pilatus6m_1_00001.cbf",
    "SPring8_ADSC_SN916/Xtal17-2phi_3_015.cbf",
    "SPring8_BL12B2_MX225HE/lys001_000001.img",
    "SPring8_BL12B2_MX225HE/lys001_000091.img",
    "SPring8_BL26B1_Raxis5/raxis5_000001.img",
    "SPring8_BL26B1_Raxis5/raxis5_000091.img",
    "SPring8_BL26B1_SaturnA200/A200_000001.img",
    "SPring8_BL26B1_SaturnA200/A200_000002.img",
    "SPring8_BL26B2_MX225/2sec_Al200um_000001.img",
    "SPring8_BL26B2_MX225/2sec_Al200um_000090.img",
    "SPring8_BL32XU/lys_00001.img",
    "SPring8_BL32XU/rayonix225_0001.img",
    "SPring8_BL32XU/rayonix225hs_0001.img",
    "SPring8_BL32XU_MX225HS/ds_000001.img",
    "SPring8_BL32XU_MX225HS/ds_000045.img",
    "SPring8_BL38B1_MX225HE/bl38b1_001.img",
    "SPring8_BL38B1_MX225HE/bl38b1_090.img",
    "SPring8_BL41XU_PILATUS3_6M/data1_000001.cbf",
    "SPring8_BL41XU_PILATUS3_6M/data1_000901.cbf",
    "SPring8_BL44XU_MX300HE/bl44xu_lys_000001.img",
    "SPring8_BL44XU_MX300HE/bl44xu_lys_000002.img",
    "SRS_101/mar225_001.img",
    "SRS_142/q4_1_001.img",
    "SSRL_bl111/mar325_1_001.mccd",
    "SSRL_bl91/q315_1_001.img",
    "Texas_A_and_M_University/lyziph6p5_01_0001.sfrm",
    "XDS/INTEGRATE.HKL",
    "XDS/XDS_ASCII.HKL",
    "XDS/XPARM.XDS",
    "dials-190/whatev1_01_00001.cbf",
    "dials-190/whatev1_01_00002.cbf",
    "dials-190/whatev1_02_00001.cbf",
    "dials-190/whatev1_02_00002.cbf",
    "dials-190/whatev1_03_00001.cbf",
    "dials-190/whatev1_03_00002.cbf",
    "johns_hopkins_raxisII/lys_001.osc",
    "saturn/lyso_00001.img",
    "xia2/merge2cbf_averaged_0001.cbf",
)

_files_with_detectorbase = sorted(
    set(_files)
    - {
        "Bruker_PHOTON_II/dan_01_0001.sfrm",
        "DLS_eBIC/nanB_4_singleframe_1.ser.bz2",
        "LCLS_cspad_nexus/idx-20130301060858401.cbf",
        "LCLS_cspad_nexus/idx-20130301060858601.cbf",
        "LCLS_cspad_nexus/idx-20130301060858701.cbf",
        "LCLS_cspad_nexus/idx-20130301060858801.cbf",
        "LCLS_jungfrau/jungfrau_multipanel.cbf",
        "SACLA_MPCCD_Cheetah/run266702-0-subset.h5",
        "XDS/INTEGRATE.HKL",
        "XDS/XDS_ASCII.HKL",
        "XDS/XPARM.XDS",
    }
)


@pytest.mark.regression
@pytest.mark.parametrize("test_image", _files_with_detectorbase)
def test_detectorbase(test_image, dials_regression):
    if not h5py and test_image.endswith((".h5", ".nxs")):
        pytest.skip("could not import 'h5py'")

    if test_image.endswith(".cbf") and not cbflib_adaptbx:
        pytest.skip("No cbflib_adaptbx: CBF Detectorbase is not available")

    test_image = str(
        Path(dials_regression).joinpath("image_examples", *test_image.split("/"))
    )
    format_instance = dxtbx.format.Registry.get_format_class_for_file(test_image)
    print("Reading", test_image)
    print("Format:", format_instance)
    assert format_instance, "no matching format class found"
    instance = format_instance(test_image)

    instance.get_detectorbase()

    imgfactory = SlipViewerImageFactory(test_image)
    imgfactory.read()
    print("  Detectorbase:", instance.detectorbase.__class__.__name__)

    try:
        print(imgfactory.rawdata.focus())
    except AttributeError:
        # Not all instances have this attribute
        print("  multireadout")

    I_raw_data = imgfactory.get_raw_data()
    if not isinstance(I_raw_data, tuple):
        I_raw_data = (I_raw_data,)

    try:
        R_raw_data = instance.get_raw_data()
    except TypeError:
        R_raw_data = instance.get_raw_data(0)

    if not isinstance(R_raw_data, tuple):
        R_raw_data = (R_raw_data,)

    # NOTE dxtbx and image factory arrays are compared here for identical values.
    for Ip, Rp in zip(I_raw_data, R_raw_data):
        assert (Ip == Rp).all_eq(True)


@pytest.mark.regression
@pytest.mark.parametrize("test_image", _files)
def test_read_image(test_image, dials_regression):
    """Test that all the regression images can be read"""

    if not h5py and test_image.endswith((".h5", ".nxs")):
        pytest.skip("could not import 'h5py'")

    if dials_regression:
        test_image = os.path.join(
            dials_regression, "image_examples", *test_image.split("/")
        )

    format_instance = dxtbx.format.Registry.get_format_class_for_file(test_image)
    print("Reading", test_image)
    print("Format:", format_instance)
    assert format_instance, "no matching format class found"
    instance = format_instance(test_image)

    # Test metadata reading
    instance.get_goniometer()
    instance.get_beam()
    instance.get_scan()
    detector = instance.get_detector()
    # From old test_dxtbx; get the image size
    if detector is not None:
        detector[0].get_image_size()

    for panel in detector:
        d_mat = scitbx.matrix.sqr(panel.get_d_matrix())
        if d_mat.determinant() <= 0:
            print("  d matrix with non-positive determinant")

    # Test reading of the raw data
    # XDS, HKL we expect to fail for this  - so skip this part for those
    if not test_image.endswith(("XDS", "HKL")):
        try:
            R_raw_data = instance.get_raw_data()
        except TypeError:
            R_raw_data = instance.get_raw_data(0)

        if not isinstance(R_raw_data, tuple):
            R_raw_data = (R_raw_data,)

        print("%-40s" % format_instance.__name__, R_raw_data[0].focus())

        # Specific test for cctbx/dxtbx#163. This test will fail if char is unsigned.
        if "APS_24IDC" in test_image and "pilatus_1_0001.cbf" in test_image:
            d = R_raw_data[0]
            assert (
                flex.sum(d.as_1d().select(d.as_1d() >= 0)) == 20108255
            )  # matches total counts from dxtbx.print_header


@pytest.mark.parametrize("test_image", _files)
def test_format_class_API_assumptions(test_image, dials_regression):
    """For a given image file, walk the whole DAG of Format objects, and
    verify the following basic assumptions for format classes:
    * Any file must only be understood by a single leaf node format class.
    * No .understand() call on any top level format class or a child class
      of another understanding format is allowed to throw an exception.
    """
    if not h5py and test_image.endswith((".h5", ".nxs")):
        pytest.skip("could not import 'h5py'")

    if dials_regression:
        test_image = os.path.join(
            dials_regression, "image_examples", *test_image.split("/")
        )

    dag = dxtbx.format.Registry.get_format_class_dag()

    def recurse(parentformat, filename, level=0):
        known_format_class = None
        multiple_formats = False
        for subformat in dag.get(parentformat, []):
            format_class = dxtbx.format.Registry.get_format_class_for(subformat)
            if not get_url_scheme(filename) in format_class.schemes:
                print("Not matching ", filename, "to", format_class)
                continue
            understood = format_class.understand(filename)
            print("%s%s: %s" % ("  " * level, subformat, understood))
            if understood:
                recursive_format_class, subtree_multiple = recurse(
                    subformat, filename, level + 1
                )
                understood_format_class = recursive_format_class or subformat
                if known_format_class and known_format_class != understood_format_class:
                    print(
                        f"File can be understood as {known_format_class} and {understood_format_class}"
                    )
                    multiple_formats = True
                known_format_class = understood_format_class
                multiple_formats |= subtree_multiple
        return known_format_class, multiple_formats

    understood_format, multiple_formats = recurse("Format", test_image)

    assert not multiple_formats, "image file understood by multiple Format objects"
    # It's a failure if nothing could understand this file
    assert understood_format, "No formatter could be found"
    print("File understood as", understood_format)


@pytest.mark.regression
@pytest.mark.parametrize(
    "image_to_test",
    # Test miniCBF and a flavour of FullCBF
    ["DLS_I02/X4_wide_M1S4_1_0001.cbf", "DLS_I04/grid_full_cbf_0005.cbf"],
)
@pytest.mark.parametrize(
    "compression,extension",
    [(gzip.GzipFile, "gz"), (bz2.BZ2File, "bz2")],
    ids=["gzip", "bz2"],
)
def test_compressed_images(
    dials_regression, compression, extension, image_to_test, tmp_path
):
    path = Path(dials_regression) / "image_examples" / image_to_test
    if not path.exists():
        pytest.skip(str(path) + " not present in dials_regression")

    # Compress this file and write to disk
    test_image_for_reading = str(tmp_path / (path.name + "." + extension))
    with path.open("rb") as f_in:
        with compression(test_image_for_reading, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)

    # Find and instantiate the format class
    format_class = dxtbx.format.Registry.get_format_class_for_file(
        test_image_for_reading
    )
    assert format_class is not None, "no matching format class found"
    instance = format_class(test_image_for_reading)

    # Test metadata reading
    assert instance.get_goniometer()
    assert instance.get_beam()
    assert instance.get_scan()
    assert instance.get_detector()
    # Test data reading
    assert instance.get_raw_data()



from __future__ import annotations

import ast
import optparse
import os
import shutil
import subprocess
import sys
from pathlib import Path
from urllib.request import urlretrieve

import dxtbx.util


def find_format_classes(
    directory: Path, base_python_path: str = "dxtbx.format"
) -> list[str]:
    format_classes = []
    for name in directory.glob("Format*.py"):
        content = name.read_bytes()
        try:
            parsetree = ast.parse(content)
        except SyntaxError:
            print(f"  *** Could not parse {name}")
            continue
        for top_level_def in parsetree.body:
            if not isinstance(top_level_def, ast.ClassDef):
                continue
            base_names = [
                baseclass.id
                for baseclass in top_level_def.bases
                if isinstance(baseclass, ast.Name)
            ]
            if any(n.startswith("Format") for n in base_names):
                classname = top_level_def.name
                format_classes.append(
                    "{classname}:{baseclasses} = {base_python_path}.{modulename}:{classname}".format(
                        baseclasses=",".join(base_names),
                        base_python_path=base_python_path,
                        classname=classname,
                        modulename=name.stem,
                    )
                )
                print(f"  found {classname} based on {base_names}")
    return format_classes


def _ensure_cert_file():
    """
    Validate that the SSL_CERT_FILE environment variable is correct.

    On MacOS .pkg installations, the SSL_CERT_FILE path is incorrect.
    This means that the following GET fails. Check for this, and work
    out where certifi actually is.
    """
    cert_file = os.getenv("SSL_CERT_FILE")
    if cert_file and not Path(cert_file).resolve().is_file():
        try:
            import certifi

            os.environ["SSL_CERT_FILE"] = certifi.where()
        except ModuleNotFoundError:
            # Better to try default instead of a nonexistent file
            del os.environ["SSL_CERT_FILE"]


def install_package(home_location: Path, format_classes: list[str]):
    setup = (
        """
from __future__ import absolute_import, division, print_function

import os
import sys
from setuptools import setup

setup(
    author="Markus Gerstel",
    author_email="scientificsoftware@diamond.ac.uk",
    entry_points={
        "dxtbx.format": """
        + repr(format_classes)
        + """,
    },
    include_package_data=True,
    license="BSD license",
    name="dxtbx_custom",
    packages=["dxtbx_custom"],
    version="1.0.0",
    zip_safe=False,
)
"""
    )

    (home_location / "setup.py").write_text(setup)
    (home_location / "__init__.py").touch()
    custom_folder = home_location / "dxtbx_custom"
    if not custom_folder.exists():
        os.symlink(home_location, custom_folder)
    subprocess.run(
        ["libtbx.python", str(home_location / "setup.py"), "develop", "--user"],
        cwd=home_location,
        capture_output=True,
        check=True,
    )


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = optparse.OptionParser(
        usage="dxtbx.install_format (--user | --global) [/path/to/format/class.py] [URL]",
        description=(
            "Updates the dxtbx format class registry and installs format classes "
            "by copying/downloading them into the format class directory. "
            "The command must be re-run whenever format classes are added or removed "
            "manually."
        ),
    )
    parser.add_option("-?", action="help", help=optparse.SUPPRESS_HELP)
    parser.add_option(
        "-g",
        "--global",
        action="store_true",
        dest="glob",
        default=False,
        help=(
            "Install format classes globally (requires write access to base python, "
            "affects everyone using the installation, files go to /build/dxtbx/formats)"
        ),
    )
    parser.add_option(
        "-u",
        "--user",
        action="store_true",
        dest="user",
        default=False,
        help=(
            "Install format classes for current user only (requires write access to home directory, "
            "affects all python installations for the current user, files go to ~/.dxtbx)"
        ),
    )
    options, args = parser.parse_args(args)

    if options.glob:
        import libtbx.load_env

        home_location = Path(abs(libtbx.env.build_path)) / "dxtbx" / "formats"
    elif options.user:
        home_location = Path.home() / ".dxtbx"
    else:
        parser.print_help()
        print("\nYou must specify --global or --user")
        sys.exit(1 if args else 0)
    home_location.mkdir(exist_ok=True, parents=True)

    for fc in args:
        try:
            local_file = Path(fc)
        except Exception:
            local_file = False
        if local_file and local_file.is_file():
            home_location_copy = home_location / local_file.name
            if local_file == home_location_copy:
                continue
            shutil.copy(local_file, home_location_copy)
            print(f"Copied {local_file} to {home_location_copy}")
            continue
        # Download the file from `url` and save it locally under `file_name`:
        if "/" in fc:
            _ensure_cert_file()
            local_file = home_location / (Path(fc.split("/")[-1]).stem + ".py")
            try:
                urlretrieve(fc, str(local_file))
            except Exception as e:
                print("Could not load from URL:", e)
                local_file = False
            if local_file and local_file.is_file():
                print(f"Downloaded {fc} to {local_file}")
                continue
        sys.exit("Could not understand " + repr(fc))

    format_classes = find_format_classes(home_location, base_python_path="dxtbx_custom")
    install_package(home_location, format_classes)
    print("\nFormat classes installed, format class registry updated.")


if __name__ == "__main__":
    run()



from __future__ import annotations

import copy
import pickle
import random

import pytest

from cctbx.eltbx import attenuation_coefficient
from libtbx.test_utils import approx_equal
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.model import Beam, Detector, Panel, ParallaxCorrectedPxMmStrategy
from dxtbx.model.detector_helpers import (
    get_detector_projection_2d_axes,
    get_panel_projection_2d_from_axes,
    set_mosflm_beam_centre,
)
from dxtbx.model.experiment_list import ExperimentListFactory


def create_detector(offset):
    # Create the detector
    detector = Detector(
        Panel(
            "",  # Type
            "Panel",  # Name
            (10, 0, 0),  # Fast axis
            (0, 10, 0),  # Slow axis
            (0 + offset, 0 + offset, 200 - offset),
            # Origin
            (0.172, 0.172),  # Pixel size
            (512, 512),  # Image size
            (0, 1000),  # Trusted range
            0.1,  # Thickness
            "Si",  # Material
            identifier="123",
        )
    )
    return detector


def create_multipanel_detector(offset, ncols=3, nrows=2):
    reference_detector = create_detector(offset)
    reference_panel = reference_detector[0]
    panel_npix = reference_panel.get_image_size()
    panel_npix = int(panel_npix[0] / ncols), int(panel_npix[1] / nrows)

    multi_panel_detector = Detector()
    for i in range(ncols):
        for j in range(nrows):
            origin_pix = i * panel_npix[0], j * panel_npix[1]
            origin = reference_panel.get_pixel_lab_coord(origin_pix)
            new_panel = Panel(
                reference_panel.get_type(),
                reference_panel.get_name() + str(j + i * nrows),
                reference_panel.get_fast_axis(),
                reference_panel.get_slow_axis(),
                origin,
                reference_panel.get_pixel_size(),
                panel_npix,
                reference_panel.get_trusted_range(),
                reference_panel.get_thickness(),
                reference_panel.get_material(),
                identifier=reference_panel.get_identifier(),
            )
            multi_panel_detector.add_panel(new_panel)

    return multi_panel_detector


@pytest.fixture
def detector():
    return create_detector(offset=0)


def test_get_gain(detector):
    detector[0].set_gain(2.0)
    assert abs(detector[0].get_gain() - 2.0) < 1e-7


def test_get_pedestal(detector):
    detector[0].set_pedestal(2.0)
    assert abs(detector[0].get_pedestal() - 2.0) < 1e-7


def test_get_identifier(detector):
    detector[0].set_identifier("HELLO")
    assert detector[0].get_identifier() == "HELLO"
    detector2 = pickle.loads(pickle.dumps(detector))
    assert detector[0].get_identifier() == detector2[0].get_identifier()


def test_get_pixel_lab_coord(detector):
    eps = 1e-7

    # Check lab coordinates at the origin
    orig = detector[0].get_pixel_lab_coord((0, 0))
    dorig = abs(matrix.col(orig) - matrix.col(detector[0].get_origin()))
    assert dorig < eps

    # Check lab coordinate at the opposite corner
    corner = detector[0].get_pixel_lab_coord((512, 512))
    dcorner = abs(matrix.col(corner) - matrix.col(corner))
    assert dcorner < eps


def test_get_image_size_mm(detector):
    eps = 1e-7
    size = detector[0].get_image_size_mm()
    size2 = (512 * 0.172, 512 * 0.172)
    dsize = abs(matrix.col(size) - matrix.col(size2))
    assert dsize < eps


def test_is_value_in_trusted_range(detector):
    """Check values are either inside or outside trusted range. Note trusted_range
    is defined as the inclusive [min-valid-value, max-valid-value], which in
    this test has been set to [0,1000]"""
    assert detector[0].is_value_in_trusted_range(-1) is False
    assert detector[0].is_value_in_trusted_range(0) is True
    assert detector[0].is_value_in_trusted_range(1000) is True
    assert detector[0].is_value_in_trusted_range(1001) is False


def test_is_coord_valid(detector):
    """Check points are either inside or outside detector range."""
    assert detector[0].is_coord_valid((-1, 256)) is False
    assert detector[0].is_coord_valid((256, 256)) is True
    assert detector[0].is_coord_valid((512, 256)) is False
    assert detector[0].is_coord_valid((256, -1)) is False
    assert detector[0].is_coord_valid((256, 256)) is True
    assert detector[0].is_coord_valid((256, 513)) is False


def test_pixel_to_millimeter_to_pixel(detector):
    eps = 1e-7

    # Pick some random pixels and check that px -> mm -> px give px == px
    w, h = detector[0].get_image_size()
    pixels = flex.vec2_double(
        (random.random() * w, random.random() * h) for i in range(100)
    )
    xy_mm = detector[0].pixel_to_millimeter(pixels)
    xy_px = detector[0].millimeter_to_pixel(xy_mm)
    assert approx_equal(xy_px, pixels, eps=eps)
    for xy in pixels:
        xy_mm = detector[0].pixel_to_millimeter(xy)
        xy_px = detector[0].millimeter_to_pixel(xy_mm)
        assert abs(matrix.col(xy_px) - matrix.col(xy)) < eps


def test_parallax_correction():
    # Attenuation length
    table = attenuation_coefficient.get_table("Si")
    mu = table.mu_at_angstrom(1) / 10.0
    t0 = 0.320

    # Create the detector
    detector = Detector(
        Panel(
            "",  # Type
            "",  # Name
            (10, 0, 0),  # Fast axis
            (0, 10, 0),  # Slow axis
            (0, 0, 200),  # Origin
            (0.172, 0.172),  # Pixel size
            (512, 512),  # Image size
            (0, 1000),  # Trusted range
            0.0,  # Thickness
            "",  # Material
            ParallaxCorrectedPxMmStrategy(mu, t0),
        )
    )
    for i in range(10000):
        mm = (random.uniform(-1000, 1000), random.uniform(-1000, 1000))
        px = detector[0].millimeter_to_pixel(mm)
        mm2 = detector[0].pixel_to_millimeter(px)
        assert abs(matrix.col(mm) - matrix.col(mm2)) < 1e-3


def test_get_names(detector):
    names = detector.get_names()
    assert len(names) == 1
    assert names[0] == "Panel"


def test_get_thickness(detector):
    for panel in detector:
        assert panel.get_thickness() == 0.1


def test_get_material(detector):
    for panel in detector:
        assert panel.get_material() == "Si"


def test_set_mosflm_beam_centre(detector):
    wavelength = 1.0
    panel = detector[0]
    detector_normal = matrix.col(panel.get_normal())
    _ = matrix.col(panel.get_origin())
    _ = matrix.col(panel.get_fast_axis())
    _ = matrix.col(panel.get_slow_axis())
    _ = panel.get_image_size_mm()

    s0 = (1.0 / wavelength) * detector_normal
    beam = Beam(-s0.normalize(), wavelength)

    beam_centre = matrix.col(panel.get_beam_centre(beam.get_s0()))
    origin_shift = matrix.col((1, 0.5))
    new_beam_centre = beam_centre + origin_shift

    new_mosflm_beam_centre = tuple(reversed(new_beam_centre))

    set_mosflm_beam_centre(detector, beam, new_mosflm_beam_centre)

    assert (
        matrix.col(panel.get_beam_centre(beam.get_s0()))
        - matrix.col(tuple(reversed(new_mosflm_beam_centre)))
    ).length() < 1e-6

    # test resolution methods
    beam = Beam(direction=(0, 0, 1), wavelength=1.0)
    d_min1 = detector.get_max_resolution(beam.get_s0())
    d_min2 = detector.get_max_inscribed_resolution(beam.get_s0())
    assert d_min1 < d_min2


def test_panel_mask():
    panel = Panel()
    panel.set_image_size((100, 100))
    panel.add_mask(40, 0, 60, 100)
    panel.add_mask(0, 40, 100, 60)
    panel.set_trusted_range((0, 9))

    data = flex.double(flex.grid(100, 100))
    data[10, 10] = -1
    data[20, 20] = 10
    data[30, 30] = 100
    data[40, 40] = -10

    m1 = panel.get_untrusted_rectangle_mask()
    m2 = panel.get_trusted_range_mask(data)

    count = 0
    for j in range(100):
        for i in range(40, 60):
            assert m1[j, i] is False
            count += 1
    for i in range(100):
        for j in range(40, 60):
            if i >= 40 and i < 60:
                continue
            assert m1[j, i] is False
            count += 1
    assert m1.count(False) == count, "%d, %d" % (m1.count(False), count)

    assert m2.count(False) == 4
    assert m2[10, 10] is False
    assert m2[20, 20] is False
    assert m2[30, 30] is False
    assert m2[40, 40] is False


def test_equality():
    detector = create_detector(offset=0)

    # Create another detector with different origin
    # Equality operator on detector objects must find differences in origin
    detector_moved = create_detector(offset=100)
    assert detector != detector_moved

    # Equality operator on detector objects must identify identical detectors
    detector_moved_copy = create_detector(offset=100)
    assert detector_moved == detector_moved_copy


def test_panel_equality():
    panel = create_detector(offset=0)[0]
    panel2 = copy.deepcopy(panel)
    assert panel == panel2

    panel2.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(1, 1))
    assert panel != panel2

    panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(1, 1))
    assert panel == panel2


def test_get_detector_projection_2d_axes():
    # The function get_detector_projection_2d_axes should give the same results even if the
    # detector is rotated in the laboratory frame

    # Use a multipanel detector
    detector = create_multipanel_detector(offset=0)

    # Get 2D origin, fast and slow vectors for the detector
    o, f, s = get_detector_projection_2d_axes(detector)

    # Now rotate the detector by 30 degrees around an arbitrary axis
    h = detector.hierarchy()
    fast = matrix.col(h.get_fast_axis())
    slow = matrix.col(h.get_slow_axis())
    origin = matrix.col(h.get_origin())

    axis = matrix.col(flex.random_double_point_on_sphere())
    rot = axis.axis_and_angle_as_r3_rotation_matrix(30, deg=True)
    for panel in detector:
        fast = matrix.col(panel.get_fast_axis())
        slow = matrix.col(panel.get_slow_axis())
        origin = matrix.col(panel.get_origin())
        panel.set_frame(rot * fast, rot * slow, rot * origin)

    # Get 2D origin, fast and slow vectors for the rotated detector
    new_o, new_f, new_s = get_detector_projection_2d_axes(detector)

    for o1, o2 in zip(o, new_o):
        assert o1 == pytest.approx(o2)
    for f1, f2 in zip(f, new_f):
        assert f1 == pytest.approx(f2)
    for s1, s2 in zip(s, new_s):
        assert s1 == pytest.approx(s2)


def test_get_panel_projection_2d_from_axes(dials_data):

    # Get test data
    pytest.importorskip("h5py")
    filename = dials_data("image_examples", pathlib=True) / "dectris_eiger_master.h5"
    experiment = ExperimentListFactory.from_filenames([filename])[0]
    detector = experiment.detector

    # Get 2d axes
    origins, fast_axes, slow_axes = get_detector_projection_2d_axes(detector)

    # Get panel 0 data
    panel = detector[0]
    image_data = experiment.imageset.get_raw_data(0)[0]
    fast_axis = matrix.col(fast_axes[0] + (0,))
    slow_axis = matrix.col(slow_axes[0] + (0,))
    origin = matrix.col(origins[0] + (0,)) * 1e-3

    # Get 2d projection
    rotation, translation = get_panel_projection_2d_from_axes(
        panel, image_data, fast_axis, slow_axis, origin
    )

    expected_rotation = (
        1.0,
        0.0,
        0.0,
        1.0,
    )
    expected_translation = (1634.5, 1555.0)

    assert len(rotation) == len(expected_rotation)
    for rot, expected in zip(rotation, expected_rotation):
        assert rot == pytest.approx(expected)

    assert len(translation) == len(expected_translation)
    for t, expected in zip(translation, expected_translation):
        assert t == pytest.approx(expected)


def test_panel_get_projection_2d():

    detector = create_detector(offset=0)
    panel = detector[0]

    # Valid projection values
    valid_rotation = (1, 0, 0, 1)
    valid_translation = detector[0].get_image_size()

    # Test panel with no projections set explicitly has no 2d projection
    assert not panel.get_projection_2d()

    # Test panel with non-empty 2d projection has 2d projection
    panel.set_projection_2d(valid_rotation, valid_translation)
    assert panel.get_projection_2d()

    # Test values are set correctly
    rotation, translation = panel.get_projection_2d()
    assert rotation == valid_rotation
    assert translation == valid_translation


def test_detector_has_projection_2d():

    # Valid rotation value
    rotation = (1, 0, 0, 1)

    ## Test single panel detector
    detector = create_detector(offset=0)

    # Test detector with no projections set explicity has no 2d projection
    assert not detector.has_projection_2d()

    # Test detector with all panels having 2d projections gives a detector
    # with a 2d projection
    detector2 = create_detector(offset=0)
    image_size = detector2[0].get_image_size()
    for i in detector2:
        i.set_projection_2d(rotation, (image_size[0], 0))

    assert detector2.has_projection_2d()
    assert detector != detector2

    ## Test multipanel detector
    detector = create_multipanel_detector(offset=0)

    # Test detector with no projections set explicity has no 2d projection
    assert not detector.has_projection_2d()

    detector2 = create_multipanel_detector(offset=0)
    image_size = detector2[0].get_image_size()
    for i in detector2:
        # Test detector with only some panels with 2d projections gives a
        # detector without a 2d projection
        assert not detector2.has_projection_2d()
        i.set_projection_2d(rotation, (image_size[0], 0))

    # Test detector with all panels having 2d projections gives a detector
    # with a 2d projection
    assert detector2.has_projection_2d()
    assert detector != detector2


def test_pickle_suite():

    rotation = (1, 0, 0, 1)

    ## Test single panel detector without 2d projection
    detector = create_detector(offset=0)
    detector2 = pickle.loads(pickle.dumps(detector))
    assert detector == detector2

    ## Test single panel detector with 2d projection
    image_size = detector[0].get_image_size()
    for i in detector:
        i.set_projection_2d(rotation, (image_size[0], 0))

    detector2 = pickle.loads(pickle.dumps(detector))
    assert detector == detector2

    ## Test multipanel detector without 2d projection
    detector = create_multipanel_detector(offset=0)
    detector2 = pickle.loads(pickle.dumps(detector))
    assert detector == detector2

    ## Test multipanel detector with 2d projection
    image_size = detector[0].get_image_size()
    for i in detector:
        i.set_projection_2d(rotation, (image_size[0], 0))

    detector2 = pickle.loads(pickle.dumps(detector))
    assert detector == detector2



from __future__ import annotations

import logging
import os
import sys
import typing

import libtbx.load_env

import dxtbx.format.Registry

from .version import version as __version__  # noqa: F401

if typing.TYPE_CHECKING:
    from dxtbx.format.Format import Format

# Ensures that HDF5 has the conda_base plugin path configured.
#
# Ideally this will be properly configured by the conda environment.
# However, currently the dials-installer will not install a path-correct
# conda_base folder, so it needs to be updated manually.

_hdf5_plugin_path = libtbx.env.under_base(os.path.join("lib", "hdf5", "plugin"))

# Inject via the environment if h5py not used yet, or else use h5py
if "h5py" not in sys.modules:
    os.environ["HDF5_PLUGIN_PATH"] = (
        _hdf5_plugin_path + os.pathsep + os.getenv("HDF5_PLUGIN_PATH", "")
    )
else:
    # We've already loaded h5py, so setting the environment variable won't work
    # We need to use the h5py API to add a plugin path
    import h5py

    h5_plugin_paths = [h5py.h5pl.get(i).decode() for i in range(h5py.h5pl.size())]
    if _hdf5_plugin_path not in h5_plugin_paths:
        h5py.h5pl.prepend(_hdf5_plugin_path.encode())

logging.getLogger("dxtbx").addHandler(logging.NullHandler())


class IncorrectFormatError(RuntimeError):
    """
    An exception class for an incorrect format error
    """

    def __init__(self, format_instance, filename):
        super().__init__(
            "Could not open {} as {}".format(filename, str(format_instance))
        )
        self.args = (format_instance, filename)


def load(filename: str | bytes | os.PathLike) -> Format:
    """Use DXTBX to load the input filename.

    Args:
        filename:  The input filename

    Returns
        A dxtbx Format-subclass instance for the file type

    Raises:
        IOError: If the file format could not be determined
    """
    format_instance = dxtbx.format.Registry.get_format_class_for_file(filename)
    if not format_instance:
        raise ValueError(f"Could not find matching format for {filename!r}")
    return format_instance(os.fspath(filename))



from __future__ import annotations

import dxtbx.model


def test_scan_to_string_does_not_crash_on_empty_scan():
    print(dxtbx.model.Scan())


def test_scan_wrap_around_zero():
    filenames = [f"foobar_{n:03d}.cbf" for n in range(350, 370)]
    starts = [r % 360 for r in range(350, 370)]
    ends = [(r + 1) % 360 for r in range(350, 370)]
    scans = [
        dxtbx.model.scan.ScanFactory.single_file(f, [0.1], s, e - s, 0)
        for f, s, e in zip(filenames, starts, ends)
    ]
    s0 = scans[0]
    for s in scans[1:]:
        assert s.get_oscillation()[1] == 1
        s0 += s
    assert s0.get_oscillation() == (350, 1)
    assert s0.get_image_range() == (350, 369)



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1

from __future__ import annotations

import os
import sys
from collections.abc import Sequence

import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d

from libtbx.phil import parse
from libtbx.utils import Sorry
from scitbx.matrix import col

import dxtbx.util
from dxtbx.model.detector_helpers import get_detector_projection_2d_axes
from dxtbx.model.experiment_list import ExperimentListFactory

usage = """Plot dxtbx detector models. Provide multiple json files if desired
Example: dxtbx.plot_detector_models experiments1.json experiments2.json
"""


phil_scope = parse(
    """
  show_origin_vectors = True
    .type = bool
    .help = If true, draw origin vectors as arrows
  orthographic = False
    .type = bool
    .help = If true, draw an orthographic projection (IE drop the Z-axis)
  project_onto = *lab_xy image_plane
    .type = choice
    .help = "If doing an orthographic projection, choose whether to project"
            "onto the laboratory frame or a best-fit plane to the panels"
  panel_numbers = True
    .type = bool
    .help = If true, label panel numbers
  pdf_file = None
    .type = path
    .help = If not None, save the result as a pdf figure.
  plot_all_detectors = True
    .type = bool
    .help = If False, plot only the first detector model found
"""
)


# http://stackoverflow.com/questions/22867620/putting-arrowheads-on-vectors-in-matplotlibs-3d-plot
class Arrow3D(FancyArrowPatch):
    def __init__(self, xs, ys, zs, *args, **kwargs):
        super().__init__((0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def do_3d_projection(self, renderer=None):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))

        return np.min(zs)


def plot_group(
    g, color, ax, orthographic=False, show_origin_vectors=True, panel_numbers=True
):
    panel_corners = []
    # recursively plot a detector group
    p = g.parent()
    if show_origin_vectors:
        if p is None:
            # parent origin
            pori = (0, 0, 0)
        else:
            # parent origin
            pori = p.get_origin()
        ori = g.get_origin()
        if not orthographic:
            a = Arrow3D(
                [pori[0], ori[0]],
                [pori[1], ori[1]],
                [pori[2], ori[2]],
                mutation_scale=20,
                lw=1,
                arrowstyle="-|>",
                color="gray",
            )
            ax.add_artist(a)
    if g.is_group():
        for c in g:
            # plot all the children
            panel_corners.append(
                plot_group(
                    c, color, ax, orthographic, show_origin_vectors, panel_numbers
                )
            )
    else:
        # plot the panel boundaries
        size = g.get_image_size()
        p0 = col(g.get_pixel_lab_coord((0, 0)))
        p1 = col(g.get_pixel_lab_coord((size[0] - 1, 0)))
        p2 = col(g.get_pixel_lab_coord((size[0] - 1, size[1] - 1)))
        p3 = col(g.get_pixel_lab_coord((0, size[1] - 1)))
        v1 = p1 - p0
        v2 = p3 - p0
        vcen = ((v2 / 2) + (v1 / 2)) + p0
        z = list(zip(p0, p1, p2, p3, p0))
        panel_corners.append((p0.elems, p1.elems, p2.elems, p3.elems))

        if orthographic:
            ax.plot(z[0], z[1], color=color)

            if panel_numbers:
                # Annotate with panel numbers
                ax.text(vcen[0], vcen[1], "%d" % g.index())
        else:
            ax.plot(z[0], z[1], z[2], color=color)

            if panel_numbers:
                # Annotate with panel numbers
                ax.text(vcen[0], vcen[1], vcen[2], "%d" % g.index())
    return panel_corners


def plot_image_plane_projection(detector, color, ax, panel_numbers=True):

    origin_2d, fast_2d, slow_2d = get_detector_projection_2d_axes(detector)

    for panel, origin, fast, slow in zip(detector, origin_2d, fast_2d, slow_2d):
        # plot the panel boundaries
        panel_size = panel.get_image_size_mm()
        p0 = col(origin)
        p1 = p0 + panel_size[0] * col(fast)
        p2 = p0 + panel_size[0] * col(fast) + panel_size[1] * col(slow)
        p3 = p0 + panel_size[1] * col(slow)
        v1 = p1 - p0
        v2 = p3 - p0
        vcen = ((v2 / 2) + (v1 / 2)) + p0
        z = list(zip(p0, p1, p2, p3, p0))

        ax.plot(z[0], z[1], color=color)

        if panel_numbers:
            # Annotate with panel numbers
            ax.text(vcen[0], vcen[1], "%d" % panel.index())


def flatten(xs):
    for x in xs:
        if isinstance(x, Sequence) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    args = args or sys.argv[1:]
    user_phil = []
    files = []
    for arg in args:
        if os.path.isfile(arg):
            files.append(arg)
        else:
            try:
                user_phil.append(parse(arg))
            except Exception:
                raise Sorry("Unrecognized argument %s" % arg)
    params = phil_scope.fetch(sources=user_phil).extract()
    if params.pdf_file:
        import matplotlib

        matplotlib.use("Agg")

    import matplotlib.pyplot as plt

    fig = plt.figure()
    colormap = plt.cm.gist_ncar
    colors = [colormap(i) for i in np.linspace(0, 0.9, len(files))]
    min_z = max_z = None
    ax = None
    for file_name, color in zip(files, colors):

        # read the data and get the detector models
        try:
            experiments = ExperimentListFactory.from_json_file(
                file_name, check_format=False
            )
        except ValueError:
            experiments = ExperimentListFactory.from_filenames([file_name])
        detectors = experiments.detectors()
        if not params.plot_all_detectors:
            detectors = detectors[0:1]
        for detector in detectors:
            # plot the hierarchy
            if ax is None and params.orthographic:
                ax = fig.gca()
            elif ax is None:
                ax = fig.add_subplot(projection="3d")

            if params.orthographic and params.project_onto == "image_plane":
                plot_image_plane_projection(detector, color, ax, params.panel_numbers)
            else:
                panel_corners = plot_group(
                    detector.hierarchy(),
                    color,
                    ax,
                    orthographic=params.orthographic,
                    show_origin_vectors=params.show_origin_vectors,
                    panel_numbers=params.panel_numbers,
                )

                if not params.orthographic:
                    # flatten the nested list of panel coordinates and then select
                    # all the z coordinates (i.e. every 3rd value)
                    all_z = list(flatten(panel_corners))[2::3]
                    if min_z is None:
                        min_z = min(all_z)
                        max_z = max(all_z)
                    else:
                        min_z = min(min_z, min(all_z))
                        max_z = max(max_z, max(all_z))

    plt.xlabel("x")
    plt.ylabel("y")
    if params.orthographic:
        ax.set_aspect("equal", "datalim")
    else:
        if min_z == max_z:
            min_z, max_z = plt.gca().zaxis.get_major_locator().nonsingular(min_z, max_z)
        plt.gca().set_zlim(zmin=min_z, zmax=max_z)

    if params.pdf_file:
        pp = PdfPages(params.pdf_file)
        for i in plt.get_fignums():
            pp.savefig(plt.figure(i))
        pp.close()
    else:
        plt.show()


if __name__ == "__main__":
    run()



"""
An implementation of the CBF image reader for Pilatus images, from the Pilatus
6M SN 119 currently on Diamond I24.
"""


from __future__ import annotations

import sys

from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus


class FormatCBFMiniPilatusDLS6MSN119(FormatCBFMiniPilatus):
    """A class for reading mini CBF format Pilatus images for 6M SN 119 @ DLS."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Pilatus mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniPilatus.get_cbf_header(image_file)

        year = 0

        for record in header.split("\n"):
            if "# 20" in record:
                year = int(record.replace("-", " ").replace("/", " ").split()[1])
                break

        if year <= 0:
            return False

        for record in header.split("\n"):
            if (
                "# Detector" in record
                and "PILATUS" in record
                and "S/N 60-0119" in header
            ):
                if year >= 2015:
                    return True
                else:
                    return False

        return False

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header, though for miniCBF
        there are limited options for this."""

        # FIXME this is currently a mess and needs properly sorting :(

        if "+FAST" in self._cif_header_dictionary.get("Oscillation_axis"):
            # plate mode - though should be -FAST
            return self._goniometer_factory.known_axis((-1, 0, 0))
        elif "-FAST" in self._cif_header_dictionary.get("Oscillation_axis"):
            # plate mode - after the fix
            return self._goniometer_factory.known_axis((-1, 0, 0))
        return self._goniometer_factory.known_axis((0, 1, 0))


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniPilatusDLS6MSN119.understand(arg))
